import warnings
import numpy as np
import networkx as nw
import unittest
from unittest.mock import Mock
import warnings
import os

import numpy as np
from scipy.optimize._minimize import MINIMIZE_METHODS

from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    QAOADescriptor,
    PauliOp,
    Hamiltonian,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.backends import create_device
from openqaoa.optimizers import get_optimizer
from openqaoa.algorithms.qaoa.qaoa_result import QAOAResult
from openqaoa.derivatives.derivative_functions import derivative
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.derivatives.qfim import qfim
from openqaoa.problems import MinimumVertexCover

"""
Unittest based testing of custom optimizers.
"""

cost_hamiltonian_1 = Hamiltonian(
    [
        PauliOp("ZZ", (0, 1)),
        PauliOp("ZZ", (1, 2)),
        PauliOp("ZZ", (0, 3)),
        PauliOp("Z", (2,)),
        PauliOp("Z", (1,)),
    ],
    [1, 1.1, 1.5, 2, -0.8],
    0.8,
)


class TestQAOACostBaseClass(unittest.TestCase):
    def setUp(self):
        self.log = Logger(
            {
                "func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "jac_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
                "qfim_func_evals": {
                    "history_update_bool": False,
                    "best_update_string": "HighestOnly",
                },
            },
            {
                "root_nodes": ["func_evals", "jac_func_evals", "qfim_func_evals"],
                "best_update_structure": [],
            },
        )

        self.log.log_variables(
            {"func_evals": 0, "jac_func_evals": 0, "qfim_func_evals": 0}
        )

    def __backend_params(self, cost_hamil, n_qubits):
        """Helper function to create a backend and a parameters onjects for testing."""
        mixer_hamil = X_mixer_hamiltonian(n_qubits=n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        device = create_device("local", "vectorized")
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor, device)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        return backend_obj_vectorized, variate_params

    def test_saving_feature(self):
        """Test save_intermediate in OptimizeVQA"""
        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "vgd",
                "tol": 10 ** (-9),
                "jac": jac,
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
                "save_intermediate": True,
            },
        )
        vector_optimizer()

    def test_scipy_optimizers_global(self):
        "Check that final value of all scipy MINIMIZE_METHODS optimizers agrees with pre-computed optimized value."

        # Create problem instance, cost function, and gradient functions
        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )

        niter = 5
        stepsize = 0.001
        y_precomp = [
            -2.4345058914425626,
            -2.5889608823632795,
            -2.588960865651421,
            -2.5889608823632786,
            -2.5889608823632795,
            -2.588960882363273,
            -2.5889608823632786,
            0.7484726235465329,
            -2.588960882363272,
            -2.588960882363281,
            -2.5889608823632786,
            -2.5889608823632795,
            -2.5889608823632786,
            -2.5889608823632786,
        ]
        optimizer_dicts = []
        for method in MINIMIZE_METHODS:
            optimizer_dicts.append(
                {"method": method, "maxiter": niter, "tol": 10 ** (-9)}
            )

        for i, optimizer_dict in enumerate(optimizer_dicts):
            optimizer_dict["jac"] = derivative(
                backend_obj_vectorized,
                variate_params,
                self.log,
                "gradient",
                "finite_difference",
            )
            optimizer_dict["hess"] = derivative(
                backend_obj_vectorized,
                variate_params,
                self.log,
                "hessian",
                "finite_difference",
            )

            # Optimize
            vector_optimizer = get_optimizer(
                backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict
            )
            vector_optimizer()

            y_opt = vector_optimizer.qaoa_result.intermediate["cost"]

            assert np.isclose(
                y_precomp[i], y_opt[-1], rtol=1e-04, atol=1e-04
            ), f"{optimizer_dict['method']} failed the test."

    def test_gradient_optimizers_global(self):
        "Check that final value of all implemented gradient optimizers agrees with pre-computed optimized value."

        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 10
        stepsize = 0.001

        # pre-computed final optimized costs
        y_precomp = [
            -2.4212581335011456,
            -2.4246393953483825,
            -2.47312715451289,
            -2.5031221706241906,
        ]

        optimizer_dicts = []
        optimizer_dicts.append(
            {
                "method": "vgd",
                "tol": 10 ** (-9),
                "jac": "finite_difference",
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
            }
        )
        optimizer_dicts.append(
            {
                "method": "newton",
                "tol": 10 ** (-9),
                "jac": "finite_difference",
                "hess": "finite_difference",
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
            }
        )
        optimizer_dicts.append(
            {
                "method": "natural_grad_descent",
                "tol": 10 ** (-9),
                "jac": "finite_difference",
                "maxiter": niter,
                "optimizer_options": {"stepsize": 0.01},
            }
        )
        optimizer_dicts.append(
            {
                "method": "rmsprop",
                "tol": 10 ** (-9),
                "jac": "finite_difference",
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize, "decay": 0.9, "eps": 1e-07},
            }
        )

        for i, optimizer_dict in enumerate(optimizer_dicts):
            # Optimize
            vector_optimizer = get_optimizer(
                backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict
            )
            vector_optimizer()

            y_opt = vector_optimizer.qaoa_result.intermediate["cost"]

            assert np.isclose(
                y_precomp[i], y_opt[-1], rtol=1e-04, atol=1e-04
            ), f"{optimizer_dict['method']} method failed the test."

    def test_gradient_optimizers_cans(self):
        n_qubits = 10
        graph = nw.circulant_graph(n_qubits, [1])
        cost_hamil = MinimumVertexCover(graph, field=1, penalty=10).qubo.hamiltonian
        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamil, n_qubits
        )

        optimizer_list = [
            {
                "method": method,
                "jac": "param_shift",
                "maxiter": 1000,
                "optimizer_options": {
                    "stepsize": 0.0001,
                    "n_shots_min": 5,
                    "n_shots_max": 200,
                    "n_shots_budget": 1000,
                },
            }
            for method in ["cans", "icans"]
        ]

        for optimizer_dict in optimizer_list:
            # optimize qaoa
            vector_optimizer = get_optimizer(
                backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict
            )
            vector_optimizer()

            # check that the final cost is the optimal value
            assert (
                int(
                    vector_optimizer.qaoa_result.most_probable_states[
                        "bitstring_energy"
                    ]
                )
                == 5
            ), "{} failed the test.".format(optimizer_dict["method"])
            # check that the number of shots are correct
            n_shots_used = np.array(vector_optimizer.qaoa_result.n_shots)
            assert (
                np.sum(np.int0(n_shots_used < 5) + np.int8(n_shots_used > 200)) == 0
            ), "Optimizer {} did not use the correct number of shots.".format(
                optimizer_dict["method"]
            )
            # check n_shots_budget worked
            assert (
                np.sum(20 * n_shots_used[:-1].T[0] + 40 * n_shots_used[:-1].T[1]) < 1000
                and np.sum(20 * n_shots_used.T[0] + 40 * n_shots_used.T[1]) >= 1000
            ), "Optimizer {} did not use the correct number of shots.".format(
                optimizer_dict["method"]
            )

            assert (
                np.sum(n_shots_used < 5) == 0
            ), "Optimizer {} did not use the correct number of shots.".format(
                optimizer_dict["method"]
            )

    def test_gradient_descent_step(self):
        """
        Check that implemented gradient descent takes the first two steps correctly.
        """

        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "vgd",
                "tol": 10 ** (-9),
                "jac": jac,
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
            },
        )
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate["cost"][1:4]

        # Stepwise optimize
        def step(x0):
            x1 = x0 - stepsize * jac(x0)

            variate_params.update_from_raw(x1)

            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_newton_step(self):
        """
        Check that implemented Newton descent takes the first two steps correctly.
        """

        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )
        hess = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "hessian",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "newton",
                "tol": 10 ** (-9),
                "jac": jac,
                "hess": hess,
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
            },
        )
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate["cost"][1:4]

        # Stepwise optimize
        def step(x0):
            scaled_gradient = np.linalg.solve(hess(x0), jac(x0))
            x1 = x0 - stepsize * scaled_gradient
            variate_params.update_from_raw(x1)
            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_natural_gradient_descent_step(self):
        """
        Check that implemented natural gradient descent takes the first two steps correctly.
        """

        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "natural_grad_descent",
                "tol": 10 ** (-9),
                "jac": jac,
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize},
            },
        )
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate["cost"][1:4]

        # Stepwise optimize
        def step(x0):
            qfim_ = qfim(backend_obj_vectorized, variate_params, self.log)
            scaled_gradient = np.linalg.solve(qfim_(x0), jac(x0))
            x1 = x0 - stepsize * scaled_gradient
            variate_params.update_from_raw(x1)
            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_rmsprop_step(self):
        """
        Check that implemented RMSProp takes the first two steps correctly.
        """

        backend_obj_vectorized, variate_params = self.__backend_params(
            cost_hamiltonian_1, 4
        )
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(
            backend_obj_vectorized,
            variate_params,
            self.log,
            "gradient",
            "finite_difference",
            {"stepsize": grad_stepsize},
        )

        decay = 0.9
        eps = 1e-07

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "rmsprop",
                "tol": 10 ** (-9),
                "jac": jac,
                "maxiter": niter,
                "optimizer_options": {"stepsize": stepsize, "decay": decay, "eps": eps},
            },
        )

        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate["cost"][1:4]

        # Stepwise optimize
        def step(x0, sqgrad0):
            sqgrad = decay * sqgrad0 + (1 - decay) * jac(x0) ** 2
            x1 = x0 - stepsize * jac(x0) / (np.sqrt(sqgrad) + eps)
            variate_params.update_from_raw(x1)
            return [
                x1,
                np.real(backend_obj_vectorized.expectation(variate_params)),
                sqgrad0,
            ]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        sqgrad0 = jac(x0) ** 2
        [x1, y1, sqgrad1] = step(x0, sqgrad0)
        [x2, y2, sqgrad2] = step(x1, sqgrad1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_optimize_loop_crash(self):
        """
        This tests that the optimization loop doesnt crash silently.
        An Exception gets raised.
        """

        cost_hamil = Hamiltonian(
            [
                PauliOp("ZZ", (0, 1)),
                PauliOp("ZZ", (1, 2)),
                PauliOp("ZZ", (0, 3)),
                PauliOp("Z", (2,)),
                PauliOp("Z", (1,)),
            ],
            [1, 1.1, 1.5, 2, -0.8],
            0.8,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=4)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device("local", "vectorized")
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor, device)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        niter = 5

        # Optimize
        vector_optimizer = get_optimizer(
            backend_obj_vectorized,
            variate_params,
            optimizer_dict={
                "method": "nelder-mead",
                "maxiter": niter,
            },
        )
        vector_optimizer.vqa.expectation = Mock(side_effect=Exception("Error!"))
        self.assertRaises(Exception, lambda: vector_optimizer.optimize())

        # Check that QAOA Result exists
        self.assertEqual(type(vector_optimizer.qaoa_result), QAOAResult)

    @classmethod
    def tearDownClass(cls):
        output_csv = ["oq_saved_info_job_ids.csv", "oq_saved_info_param_log.csv"]
        for each_csv in output_csv:
            try:
                if os.path.exists(each_csv):
                    os.remove(each_csv)
            except Exception:
                continue


if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=PendingDeprecationWarning)
        unittest.main()



import unittest
import numpy as np
from scipy.linalg import expm
from scipy.sparse import csc_matrix, kron, diags

# RX and CHPHASE are never used
from openqaoa.backends.qaoa_vectorized import (
    QAOAvectorizedBackendSimulator,
    _permute_qubits,
    _get_perm,
    RX,
)
from openqaoa.utilities import X_mixer_hamiltonian, ring_of_disagrees
from openqaoa.qaoa_components import (
    QAOAVariationalExtendedParams,
    QAOAVariationalStandardParams,
    Hamiltonian,
    PauliOp,
    QAOADescriptor,
    create_qaoa_variational_params,
)

######################################################
# TESTS OF SIMPLE PERMUTATION AND RESHAPING OPERATIONS
######################################################


def Disagrees_SetUp(n_qubits):
    """
    Helper function for the tests below
    """

    register = range(n_qubits)
    p = 1
    cost_hamil = ring_of_disagrees(register)
    mixer_hamil = X_mixer_hamiltonian(n_qubits)

    betas = [np.pi / 8]
    gammas = [np.pi / 4]

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p)
    variational_params_std = QAOAVariationalStandardParams(
        qaoa_descriptor, betas, gammas
    )
    # Get the part of the Hamiltonian proportional to the identity

    return register, cost_hamil, qaoa_descriptor, variational_params_std


def pauli_matrix_SetUp():
    """
    Helper function for apply_gate tests.
    """

    constI = csc_matrix(np.eye(2)).toarray()
    constX = csc_matrix(np.array([[0, 1], [1, 0]])).toarray()
    constY = csc_matrix(np.array([[0, -1j], [1j, 0]])).toarray()
    constZ = csc_matrix(np.array([[1, 0], [0, -1]])).toarray()

    return constI, constX, constY, constZ


def apply_gate_problem_SetUp():
    """
    Helper function for apply_gate tests.
    """

    cost_hamil = Hamiltonian([PauliOp("ZZ", (0, 1)), PauliOp("Z", (2,))], [1, 1], 0)
    mixer_hamil = Hamiltonian(
        [PauliOp("X", (0,)), PauliOp("X", (1,)), PauliOp("X", (2,))], [1, 1, 1], 0
    )
    theta = 0  # Don't apply mixer and driver unitaries
    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variate_params = QAOAVariationalStandardParams(qaoa_descriptor, theta, theta)
    vector_backend = QAOAvectorizedBackendSimulator(qaoa_descriptor, None, None, True)
    return cost_hamil.n_qubits, vector_backend


class TestingQAOAvectorizedBackend(unittest.TestCase):
    """
    Unittest based testing of QAOACostVector
    """

    def test_permute(self):
        nqubits = 3

        arr = np.arange(2**nqubits)
        arr.shape = [2] * nqubits
        # reshaped_arr = backend._reshape_qubits(nqubits, arr)
        perm = [2, 0, 1]
        permuted_arr = _permute_qubits(arr, perm)

        expected_arr = np.array([[[0, 2], [4, 6]], [[1, 3], [5, 7]]])

        assert np.array_equal(permuted_arr, expected_arr)

    def test_get_perm(self):
        nqubits = 4

        perm1, perminv1 = _get_perm(nqubits, [1, 3])
        perm2, perminv2 = _get_perm(nqubits, [0, 2])

        perm1_expected = np.array([2, 0, 1, 3])
        perminv1_expected = np.array([1, 2, 0, 3])

        perm2_expected = np.array([3, 1, 0, 2])
        perminv2_expected = np.array([2, 1, 3, 0])

        assert np.array_equal(perm1, perm1_expected)
        assert np.array_equal(perminv1, perminv1_expected)
        assert np.array_equal(perm2, perm2_expected)
        assert np.array_equal(perminv2, perminv2_expected)

    ##########################################################
    # TESTS OF BASIC CIRCUIT OPERATIONS (SAME AS FOR PROJECTQ)
    ##########################################################

    def test_qaoa_circuit(self):
        # Test circuit with p = 1 on 3 qubits
        # Performs a round of ZZ rotations through pi, and a round of X mixer rotations through pi
        nqubits = 3
        register = [i for i in range(nqubits)]
        p = 1
        bias_qubits = []
        bias_angles = []
        pairs = [[0, 1], [0, 2], [1, 2]]
        weights = [1, 1, 1]
        pairs_angles = [np.pi]  # [[np.pi]*len(pairs)]
        mixer_angles = [np.pi]  # [[np.pi]*nqubits]

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(pairs, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, mixer_angles, pairs_angles
        )

        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        backend_vectorized.qaoa_circuit(variational_params_std)
        wf = backend_vectorized.wavefn
        wf.shape = 2**nqubits
        wf = wf / wf[0]

        expected_wf = np.array([1, 1, 1, 1, 1, 1, 1, 1])

        assert np.allclose(wf, expected_wf)

    def test_wavefunction_single_terms(self):
        # Test wavefunction and expectation values with hamiltonian object, without 2-qubit terms
        cost_hamil = Hamiltonian([PauliOp("Z", (0,)), PauliOp("Z", (1,))], [1, 1], 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        backend_obj = QAOAvectorizedBackendSimulator(qaoa_descriptor, None, None, True)

        args = [np.pi / 4, np.pi / 4]  # beta, gamma
        variate_params.update_from_raw(args)

        assert np.allclose(backend_obj.wavefunction(variate_params), [0, 0, 0, 1j])
        assert np.isclose(backend_obj.expectation(variate_params), -1)

    def test_wavefunction(self):
        nqubits = 3

        # Test circuit with p = 1 on 3 qubits
        terms = [[0, 1], [0, 2], [0]]
        weights = [1, 1, -0.5]
        register = [0, 1, 2]
        p = 1

        betas_singles = [np.pi, 0, 0]
        betas_pairs = []
        gammas_singles = [np.pi]
        gammas_pairs = [[np.pi / 2] * 2]

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_ext = QAOAVariationalExtendedParams(
            qaoa_descriptor, betas_singles, betas_pairs, gammas_singles, gammas_pairs
        )

        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        wf = backend_vectorized.wavefunction(variational_params_ext)

        cost_op1 = diags([-1, 1, 1, -1, -1, 1, 1, -1], 0, format="csc")
        cost_op2 = diags([-1, 1, -1, 1, 1, -1, 1, -1], 0, format="csc")
        cost_op3 = -1j * diags([-1, 1, -1, 1, -1, 1, -1, 1], 0, format="csc")

        # Factors of 2 needed to produce a rotation for total
        mixer = kron(RX(0), kron(RX(0), RX(-2 * np.pi)))
        # time of pi (pi-pulse) on the Bloch sphere

        input_wf = np.ones(2**nqubits) / np.sqrt(2**nqubits)
        direct_wf = -mixer @ cost_op3 @ cost_op2 @ cost_op1 @ input_wf

        expected_wf = -1j * np.array([-1, 1, 1, -1, 1, -1, -1, 1]) / (2 * np.sqrt(2))

        assert np.allclose(wf, direct_wf)
        assert np.allclose(wf, expected_wf)

    def test_execute_exp_val(self):
        n_qubits = 8
        register, cost_hamil, qaoa_descriptor, variate_params = Disagrees_SetUp(
            n_qubits
        )

        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )
        exp_val, std_dev1 = backend_vectorized.expectation_w_uncertainty(variate_params)

        # Check correct expecation value
        assert np.isclose(exp_val, -6)

        # Check standard deviation
        # Get the matrix form of the Hamiltonian (note we just keep the diagonal part) and square it
        ham_matrix = np.zeros((2 ** len(register)))
        for i, term in enumerate(cost_hamil.terms):
            out = np.real(cost_hamil.coeffs[i])
            for qubit in register:
                if qubit in term.qubit_indices:
                    out = np.kron([1, -1], out)
                else:
                    out = np.kron([1, 1], out)
            ham_matrix += out
        ham_matrix += cost_hamil.constant

        ham_matrix_sq = np.square(ham_matrix)

        # Get the wavefunction
        wf = backend_vectorized.wavefunction(variate_params)

        # Get the probabilities
        probs = [np.abs(el) ** 2 for el in wf]

        # Standard deviation
        exp_2 = np.dot(probs, ham_matrix)
        std_dev2 = np.sqrt(np.dot(probs, ham_matrix_sq) - exp_2**2)

        assert np.isclose(std_dev1, std_dev2)

    def test_cost_call(self):
        """
        testing the __call__ method of the base class.
        Only for vectorized and Qiskit Local Statevector Backends.
        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        betas = [np.pi / 8]
        gammas = [np.pi / 4]
        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        exp_vec = backend_vectorized.expectation(variational_params_std)

        assert np.isclose(exp_vec, -6)

    def test_get_wavefunction(self):
        n_qubits = 3
        terms = [[0, 1], [0, 2], [0]]
        weights = [1, 1, -0.5]
        p = 1

        betas_singles = [np.pi, 0, 0]
        betas_pairs = []
        gammas_singles = [np.pi]
        gammas_pairs = [[1 / 2 * np.pi] * 2]

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms=terms, coeffs=weights, constant=0
        )
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalExtendedParams(
            qaoa_descriptor,
            betas_singles=betas_singles,
            betas_pairs=betas_pairs,
            gammas_singles=gammas_singles,
            gammas_pairs=gammas_pairs,
        )

        backend_vectorised_statevec = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        wf_vectorised_statevec = backend_vectorised_statevec.wavefunction(
            (variational_params_std)
        )
        expected_wf = 1j * np.array([-1, 1, 1, -1, 1, -1, -1, 1]) / (2 * np.sqrt(2))

        try:
            assert np.allclose(wf_vectorised_statevec, expected_wf)
        except AssertionError:
            assert np.allclose(
                np.real(np.conjugate(wf_vectorised_statevec) * wf_vectorised_statevec),
                np.conjugate(expected_wf) * expected_wf,
            )

    def test_exact_solution(self):
        """
        NOTE:Since the implementation of exact solution is backend agnostic
            Checking it once should be okay.

        Nevertheless, for the sake of completeness it will be tested for all backend
        instances.

        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        correct_energy = -8
        correct_config = [0, 1, 0, 1, 0, 1, 0, 1]

        # The tests pass regardless of the value of betas and gammas is this correct?
        betas = [np.pi / 8]
        gammas = [np.pi / 4]

        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )
        # exact solution is defined as the property of the cost function
        energy_vec, config_vec = backend_vectorized.exact_solution

        assert np.isclose(energy_vec, correct_energy)

        config_vec = [config.tolist() for config in config_vec]

        assert correct_config in config_vec

    def test_afunction_throws_exception(self):
        # Make sure that exception is raised when Hamiltonian contains nonclassical (non-Z or ZZ terms)

        def test_nonclassical_hamiltonian_error():
            cost_hamil = Hamiltonian(
                [PauliOp("Y", (0,)), PauliOp("Z", (1,))], [1, 1], 1
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )
            backend_obj = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, True
            )

        self.assertRaises(Exception, test_nonclassical_hamiltonian_error)

    ##########################################################
    # TESTS OF APPLY GATE METHODS
    ##########################################################

    def test_apply_rx(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rx method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rx(0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constI, constX)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_ry(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_ry method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_ry(0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constI, constY)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rz(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rz method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rz(0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constI, constZ)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rxx(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rxx method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rxx(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constX, constX)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_ryy(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_ryy method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_ryy(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constY, constY)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rzz(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rzz method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rzz(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constZ, constZ)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rzx(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rzx method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rzx(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constX, constZ)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rxz(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rxz method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rzx(1, 0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constZ, constX)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rxy(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_rxy method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rxy(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constY, constX)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_ryx(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_ryx method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_rxy(1, 0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constX, constY)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_ryz(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_ryz method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_ryz(0, 1, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constZ, constY)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    def test_apply_rzy(self):
        constI, constX, constY, constZ = pauli_matrix_SetUp()

        # Result from apply_ryz method
        angles = [0.1, np.pi / 2, np.pi / 4]

        for angle in angles:
            n_qubits, vector_backend = apply_gate_problem_SetUp()
            vector_backend.apply_ryz(1, 0, angle)

            # Result from matrix multiply exponentiated gate
            wavefn = np.ones((2**n_qubits,), dtype=complex) / np.sqrt(2**n_qubits)
            unitary = expm(
                -kron(constI, kron(constY, constZ)).toarray() * angle * 1j / 2
            )
            res_wfn = np.matmul(unitary, wavefn).reshape([2] * n_qubits)

            assert np.allclose(
                vector_backend.wavefn, res_wfn
            ), f"angle = {angle} failed, {vector_backend.wavefn} != {res_wfn}"

    # ADD TESTS FOR PREPEND AND APPEND STATES BELOW
    # def test_with_init_prog_A(self):

    #     """
    #     Checks the trivial case that when we use a given set of params (params_std) as the
    #     init_prog followed by a circuit whose angles are zero (new_params), we should get the same as
    #     if we run the program with no init_prog, but where the angles are given by params_std
    #     """

    #     n_qubits = 8
    #     _,_,_, params_std = Disagrees_SetUp(n_qubits)

    #     new_params = copy(params_std)
    #     new_params.betas = [0]
    #     new_params.gammas = [0]

    #     vector_cost = QAOACostVector(params_std, init_prog=params_std)
    #     exp_val = vector_cost.execute_exp_val(new_params)

    #     # Check correct expecation value
    #     assert np.isclose(exp_val, -6)

    # def test_with_init_prog_B(self):

    #     """
    #     Set up a problem with p = 2 and solve directly
    #     Set up the same problem with an init_prog corresponding to the p=1 params,
    #     and the program to execute being the p=2 params.
    #     """

    #     n = 4
    #     p = 2
    #     register = range(n)
    #     terms = [[0, 1], [0, 2], [0, 3], [3, 0], [1, 3]]
    #     weights = [0.1, 1, -0.5, 0.3, 3]

    #     # Direct way for p=2
    #     hyperparams = HyperParams(terms,weights,p=p)
    #     stand_params = StandardParams.linear_ramp_from_hamiltonian(hyperparams)
    #     vector_cost = QAOACostVector(stand_params)
    #     wf_direct = vector_cost.wavefunction(stand_params)

    #     # Indirect way
    #     init_hyperparams = HyperParams(terms,weights,p=1)
    #     init_betas = stand_params.betas[0]
    #     init_gammas = stand_params.gammas[0]
    #     init_prog = StandardParams(init_hyperparams, (init_betas, init_gammas))

    #     betas_step2 = stand_params.betas[1]
    #     gammas_step2 = stand_params.gammas[1]
    #     params_step2 = StandardParams(init_hyperparams,(betas_step2, gammas_step2))
    #     vector_cost = QAOACostVector(params_step2, init_prog = init_prog)
    #     wf_indirect = vector_cost.wavefunction(params_step2)

    #     assert np.allclose(wf_direct, wf_indirect)

    # def test_with_init_prog_C(self):

    #     """
    #     Similar to above, but uses a random example with ExtendedParams
    #     """

    #     n = 8
    #     p = 3
    #     reg = range(n)
    #     terms, weights = zip(*random_hamiltonian(reg).items())
    #     terms, weights = list(terms),list(weights)
    #     hyperparams = HyperParams(terms,weights,p=p)
    #     abs_params = AbstractParams(hyperparams)
    #     n_gammas_singles = len(abs_params.qubits_singles)
    #     n_gammas_pairs = len(abs_params.qubits_pairs)

    #     # Direct way
    #     betas = np.random.rand(p,n)
    #     gammas_singles = np.random.rand(p,n_gammas_singles)
    #     gammas_pairs = np.random.rand(p,n_gammas_pairs)
    #     params = (betas,gammas_singles, gammas_pairs)
    #     ext_params = ExtendedParams.from_AbstractParameters(abs_params,params)
    #     vector_cost = QAOACostVector(ext_params)
    #     wf_direct = vector_cost.wavefunction(ext_params)

    #     # Indirect way: pass in the p=2 state as the init_prog to the final step
    #     init_hyperparams = HyperParams(terms,weights,p=2)
    #     init_betas = [ext_params.betas[0], ext_params.betas[1]]
    #     init_gammas_singles = [ext_params.gammas_singles[0], ext_params.gammas_singles[1]]
    #     init_gammas_pairs = [ext_params.gammas_pairs[0], ext_params.gammas_pairs[1]]
    #     init_prog = ExtendedParams(init_hyperparams, (init_betas, init_gammas_singles, init_gammas_pairs))

    #     hyperparams3 = HyperParams(terms,weights,p=1)
    #     betas_step3 = ext_params.betas[2]
    #     gammas_singles_step3 = ext_params.gammas_singles[2]
    #     gammas_pairs_step3 = ext_params.gammas_pairs[2]
    #     params_step3 = ExtendedParams(hyperparams3,(betas_step3, gammas_singles_step3, gammas_pairs_step3))
    #     vector_cost = QAOACostVector(params_step3, init_prog = init_prog)
    #     wf_indirect = vector_cost.wavefunction(params=params_step3)

    #     assert np.allclose(wf_direct, wf_indirect)


if __name__ == "__main__":
    unittest.main()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pytest
import unittest
import networkx as nw
from braket.jobs.local import LocalQuantumJob
from unittest.mock import MagicMock

from openqaoa.problems import MinimumVertexCover
from openqaoa_braket.algorithms import AWSJobs
from openqaoa.algorithms import QAOA, RQAOA
from openqaoa.backends import create_device


class TestingAwsJobs(unittest.TestCase):
    def setUp(self):
        # the input data directory opt/braket/input/data
        os.environ["AMZN_BRAKET_INPUT_DIR"] = "./src/openqaoa-braket/tests/jobs_test_input/"
        # the output directory opt/braket/model to write ob results to
        os.environ["AMZN_BRAKET_JOB_RESULTS_DIR"] = "/oq_release_tests/testing_jobs/"
        # the name of the job
        os.environ["AMZN_BRAKET_JOB_NAME"] = "oq_release_test"
        # the checkpoint directory
        os.environ["AMZN_BRAKET_CHECKPOINT_DIR"] = "oq_test_suite_checkpoint"
        # the hyperparameter
        os.environ["AMZN_BRAKET_HP_FILE"] = ""
        # the device ARN (AWS Resource Number)
        os.environ[
            "AMZN_BRAKET_DEVICE_ARN"
        ] = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
        # the output S3 bucket, as specified in the CreateJob request’s OutputDataConfig
        os.environ["AMZN_BRAKET_OUT_S3_BUCKET"] = "amazon-braket-us-east-1-oq-testing"
        # the entry point as specified in the CreateJob request’s ScriptModeConfig
        os.environ["AMZN_BRAKET_SCRIPT_ENTRY_POINT"] = ""
        # the compression type as specified in the CreateJob request’s ScriptModeConfig
        os.environ["AMZN_BRAKET_SCRIPT_COMPRESSION_TYPE"] = ""
        # the S3 location of the user’s script as specified in the CreateJob request’s ScriptModeConfig
        os.environ["AMZN_BRAKET_SCRIPT_S3_URI"] = ""
        # the S3 location where the SDK would store the task results by default for the job
        os.environ["AMZN_BRAKET_TASK_RESULTS_S3_URI"] = ""
        # the S3 location where the job results would be stored, as specified in CreateJob request’s OutputDataConfig
        os.environ["AMZN_BRAKET_JOB_RESULTS_S3_PATH"] = ""
        # the string that should be passed to CreateQuantumTask’s jobToken parameter for quantum tasks created in the job container
        # os.environ["AMZN_BRAKET_JOB_TOKEN"] = ''

        self.n_qubits = 10

        self.vc = MinimumVertexCover(
            nw.circulant_graph(self.n_qubits, [1]), field=1.0, penalty=10
        ).qubo

    def testOsEnvironAssignment(self):
        qaoa_workflow = AWSJobs(algorithm="QaoA")
        assert qaoa_workflow.algorithm == "qaoa"
        assert qaoa_workflow.input_dir == "./src/openqaoa-braket/tests/jobs_test_input/"
        assert qaoa_workflow.device.device_name == os.environ["AMZN_BRAKET_DEVICE_ARN"]

        rqaoa_workflow = AWSJobs(algorithm="rqAoa")
        assert rqaoa_workflow.algorithm == "rqaoa"
        assert rqaoa_workflow.device.device_name == os.environ["AMZN_BRAKET_DEVICE_ARN"]

    @pytest.mark.docker_aws
    def testLocalJob(self):
        """Test an end-to-end qaoa running on a local docker instance"""

        input_data_path = os.path.join(
            os.environ["AMZN_BRAKET_INPUT_DIR"], "input_data/"
        )

        # Create the qubo and the qaoa
        q = QAOA()
        q.set_classical_optimizer(maxiter="5")
        q.set_device(
            create_device("aws", "arn:aws:braket:::device/quantum-simulator/amazon/sv1")
        )

        ### The following lines are needed to fool the github actions into correctly executing q.compile() !!
        q.device.check_connection = MagicMock(return_value=True)
        q.device.qpu_connected = True
        q.device.provider_connected = True
        q.device.n_qubits = self.n_qubits
        q.device.backend_device = ""
        q.device.aws_region = "us-east-1"

        q.compile(self.vc)
        q.dump(
            file_name="openqaoa_params.json",
            file_path=input_data_path,
            prepend_id=False,
            overwrite=True,
        )

        job = LocalQuantumJob.create(
            device="arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            source_module="./src/openqaoa-braket/tests/jobs_test_input/aws_braket_source_module/openqaoa_qaoa_script.py",
            image_uri="amazon-braket-oq-dev",
            input_data={"input_data": input_data_path},
        )

        assert (job.state() == "COMPLETED") and (job.result() != None) == True

    @pytest.mark.docker_aws
    def testLocalJobRQAOA(self):
        """Test an end-to-end rqaoa running on a local docker instance"""

        input_data_path = os.path.join(
            os.environ["AMZN_BRAKET_INPUT_DIR"], "input_data/"
        )

        # Create the rqaoa
        r = RQAOA()
        r.set_rqaoa_parameters(n_cutoff=6)
        r.set_classical_optimizer(maxiter=3, save_intermediate=False)
        r.set_device(
            create_device("aws", "arn:aws:braket:::device/quantum-simulator/amazon/sv1")
        )

        ### The following lines are needed to fool the github actions into correctly executing q.compile() !!
        r.device.check_connection = MagicMock(return_value=True)
        r.device.qpu_connected = True
        r.device.provider_connected = True
        r.device.n_qubits = self.n_qubits
        r.device.backend_device = ""
        r.device.aws_region = "us-east-1"

        r.compile(self.vc)
        r.dump(
            file_name="openqaoa_params.json",
            file_path=input_data_path,
            prepend_id=False,
            overwrite=True,
        )

        job = LocalQuantumJob.create(
            device="arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            source_module="./src/openqaoa-braket/tests/jobs_test_input/aws_braket_source_module/openqaoa_rqaoa_script.py",
            image_uri="amazon-braket-oq-dev",
            input_data={"input_data": input_data_path},
        )

        assert (job.state() == "COMPLETED") and (job.result() != None) == True


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""Rotosolve gradient free optimizer"""
# pylint: disable=too-many-branches,cell-var-from-loop

from inspect import signature
import numpy as np
from scipy.optimize import brute, shgo

from openqaoa.optimizers import (
    pennylane as qml,
)  # changed from the original PennyLane code


def _brute_optimizer(fun, num_steps, bounds=None, **kwargs):
    r"""Brute force optimizer, wrapper of scipy.optimize.brute that repeats it
    ``num_steps`` times. Signature is as expected by ``RotosolveOptimizer._min_numeric``
    below, returning a scalar minimal position and the function value at that position.
    """
    Ns = kwargs.pop("Ns")
    width = bounds[0][1] - bounds[0][0]
    center = (bounds[0][1] + bounds[0][0]) / 2
    for _ in range(num_steps):
        range_ = (center - width / 2, center + width / 2)
        center, y_min, *_ = brute(
            fun, ranges=(range_,), full_output=True, Ns=Ns, **kwargs
        )
        # We only ever use this function for 1D optimization
        center = center[0]
        width /= Ns

    return center, y_min


def _shgo_optimizer(fun, **kwargs):
    r"""Wrapper for ``scipy.optimize.shgo`` (Simplicial Homology global optimizer).
    Signature is as expected by ``RotosolveOptimizer._min_numeric`` below, providing
    a scalar minimal position and the function value at that position."""
    opt_res = shgo(fun, **kwargs)
    return opt_res.x[0], opt_res.fun


def _validate_inputs(requires_grad, args, nums_frequency, spectra):
    """Checks that for each trainable argument either the number of
    frequencies or the frequency spectrum is given."""

    if not any(requires_grad.values()):
        raise ValueError(
            "Found no parameters to optimize. The parameters to optimize "
            "have to be marked as trainable."
        )
    for arg, (arg_name, _requires_grad) in zip(args, requires_grad.items()):
        if _requires_grad:
            _nums_frequency = nums_frequency.get(arg_name, {})
            _spectra = spectra.get(arg_name, {})
            all_keys = set(_nums_frequency) | set(_spectra)

            shape = qml.math.shape(arg)
            indices = np.ndindex(shape) if len(shape) > 0 else [()]
            for par_idx in indices:
                if par_idx not in all_keys:
                    raise ValueError(
                        "Neither the number of frequencies nor the frequency spectrum "
                        f"was provided for the entry {par_idx} of argument {arg_name}."
                    )


def _restrict_to_univariate(fn, arg_idx, par_idx, args, kwargs):
    r"""Restrict a function to a univariate function for given argument
    and parameter indices.

    Args:
        fn (callable): Multivariate function
        arg_idx (int): Index of the argument that contains the parameter to restrict
        par_idx (tuple[int]): Index of the parameter to restrict to within the argument
        args (tuple): Arguments at which to restrict the function.
        kwargs (dict): Keyword arguments at which to restrict the function.

    Returns:
        callable: Univariate restriction of ``fn``. That is, this callable takes
        a single float value as input and has the same return type as ``fn``.
        All arguments are set to the given ``args`` and the input value to this
        function is added to the marked parameter.
    """
    the_arg = args[arg_idx]
    if len(qml.math.shape(the_arg)) == 0:
        shift_vec = qml.math.ones_like(the_arg)
    else:
        shift_vec = qml.math.zeros_like(the_arg)
        shift_vec = qml.math.scatter_element_add(shift_vec, par_idx, 1.0)

    def _univariate_fn(x):
        return fn(
            *args[:arg_idx], the_arg + shift_vec * x, *args[arg_idx + 1 :], **kwargs
        )

    return _univariate_fn


class RotosolveOptimizer:
    r"""Rotosolve gradient-free optimizer.

    The Rotosolve optimizer minimizes an objective function with respect to the parameters of a
    quantum circuit without the need for calculating the gradient of the function. The algorithm
    updates the parameters :math:`\boldsymbol{\theta} = \theta_1, \dots, \theta_D` by
    separately reconstructing the cost function with respect to each circuit parameter,
    while keeping all other parameters fixed.

    Args:
        substep_optimizer (str or callable): Optimizer to use for the substeps of Rotosolve
            that carries out a univariate (i.e., single-parameter) global optimization.
            *Only used if there are more than one frequency for a given parameter.*
            It must take as inputs:

            - A function ``fn`` that maps scalars to scalars,

            - the (keyword) argument ``bounds``, and

            - optional keyword arguments.

            It must return two scalars:

            - The input value ``x_min`` for which ``fn`` is minimal, and

            - the minimal value ``y_min=fn(x_min)`` or ``None``.

            Alternatively, the following optimizers are built-in and can be chosen by
            passing their name:

            - ``"brute"``: An iterative version of
              `SciPy's brute force optimizer <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.brute.html>`_.
              It evaluates the function at ``Ns`` equidistant points across the range
              :math:`[-\pi, \pi]` and iteratively refines the range around the point
              with the smallest cost value for ``num_steps`` times.

            - ``"shgo"``: `SciPy's SHGO optimizer <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.shgo.html>`_.

        substep_kwargs (dict): Keyword arguments to be passed to the ``substep_optimizer``
            callable. For ``substep_optimizer="shgo"``, the original keyword arguments of
            the SciPy implementation are available, for ``substep_optimizer="brute"`` the
            keyword arguments ``ranges``, ``Ns`` and ``num_steps`` are useful.
            *Only used if there are more than one frequency for a given parameter.*

    For each parameter, a purely classical one-dimensional global optimization over the
    interval :math:`(-\pi,\pi]` is performed, which is replaced automatically by a
    closed-form expression for the optimal value if the :math:`d\text{th}` parametrized
    gate has only two eigenvalues. This means that ``substep_optimizer`` and
    ``substep_kwargs`` will not be used for these parameters.
    In this case, the optimal value :math:`\theta^*_d` is given analytically by

    .. math::

        \theta^*_d &= \underset{\theta_d}{\text{argmin}}\left<H\right>_{\theta_d}\\
              &= -\frac{\pi}{2} - \text{arctan2}\left(2\left<H\right>_{\theta_d=0}
              - \left<H\right>_{\theta_d=\pi/2} - \left<H\right>_{\theta_d=-\pi/2},
              \left<H\right>_{\theta_d=\pi/2} - \left<H\right>_{\theta_d=-\pi/2}\right),

    where :math:`\left<H\right>_{\theta_d}` is the expectation value of the objective function
    restricted to only depend on the parameter :math:`\theta_d`.

    .. warning::

        The built-in one-dimensional optimizers ``"brute"`` and ``"shgo"`` for the substeps
        of a Rotosolve optimization step use the interval :math:`(-\pi,\pi]`, rescaled with
        the inverse smallest frequency as default domain to optimize over. For complicated
        cost functions, this domain might not be suitable for the substep optimization and
        an appropriate range should be passed via ``bounds`` in ``substep_kwargs``.

    The algorithm is described in further detail in
    `Vidal and Theis (2018) <https://arxiv.org/abs/1812.06323>`_,
    `Nakanishi, Fujii and Todo (2019) <https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.043158>`_,
    `Parrish et al. (2019) <https://arxiv.org/abs/1904.03206>`_,
    and
    `Ostaszewski et al. (2019) <https://quantum-journal.org/papers/q-2021-01-28-391/>`_,
    and the reconstruction method used for more general operations is described in
    `Wierichs et al. (2022) <https://doi.org/10.22331/q-2022-03-30-677>`_.

    .. warning::

        ``RotosolveOptimizer`` will only update parameters that are *explicitly*
        marked as trainable. This can be done via ``requires_grad`` if using Autograd
        or PyTorch. ``RotosolveOptimizer`` is not yet implemented to work in a stable
        manner with TensorFlow or JAX.

    **Example:**

    Initialize the optimizer and set the number of steps to optimize over.
    Recall that the optimization with ``RotosolveOptimizer`` uses global optimization substeps
    of univariate functions. The optimization technique for these substeps can be chosen via the
    ``substep_optimizer`` and ``substep_kwargs`` keyword arguments.
    Here we use the built-in iterative version of
    `SciPy's brute force optimizer <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.brute.html>`_
    with four iterations.
    We will run Rotosolve itself for three iterations.

    >>> opt_kwargs = {"num_steps": 4}
    >>> opt = qml.optimize.RotosolveOptimizer(substep_optimizer="brute", substep_kwargs=opt_kwargs)
    >>> num_steps = 3

    Next, we create a QNode we wish to optimize:

    .. code-block :: python

        dev = qml.device('default.qubit', wires=3, shots=None)

        @qml.qnode(dev)
        def cost_function(rot_param, layer_par, crot_param, rot_weights=None, crot_weights=None):
            for i, par in enumerate(rot_param * rot_weights):
                qml.RX(par, wires=i)
            for w in dev.wires:
                qml.RX(layer_par, wires=w)
            for i, par in enumerate(crot_param*crot_weights):
                qml.CRY(par, wires=[i, (i+1)%3])
            return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1) @ qml.PauliZ(2))

    This QNode is defined simply by measuring the expectation value of the tensor
    product of ``PauliZ`` operators on all qubits.
    It takes three parameters:

    - ``rot_param`` controls three Pauli rotations with three parameters, multiplied with ``rot_weights``,
    - ``layer_par`` feeds into a layer of rotations with a single parameter, and
    - ``crot_param`` feeds three parameters, multiplied with ``crot_weights``, into
      three controlled Pauli rotations.

    We also initialize a set of parameters for all these operations, and start with
    uniform weights, i.e., all ``rot_weights`` and ``crot_weights`` are set to one.
    This means that all frequencies with which the parameters in ``rot_param`` and
    ``crot_param`` enter the QNode are integer-valued.
    The number of frequencies per parameter are summarized in ``nums_frequency``.

    .. code-block :: python

        init_param = (
            np.array([0.3, 0.2, 0.67], requires_grad=True),
            np.array(1.1, requires_grad=True),
            np.array([-0.2, 0.1, -2.5], requires_grad=True),
        )
        rot_weights = np.ones(3)
        crot_weights = np.ones(3)

        nums_frequency = {
            "rot_param": {(0,): 1, (1,): 1, (2,): 1},
            "layer_par": {(): 3},
            "crot_param": {(0,): 2, (1,): 2, (2,): 2},
        }

    The keyword argument ``requires_grad`` can be used to determine whether the respective
    parameter should be optimized or not, following the behaviour of gradient computations and
    gradient-based optimizers when using Autograd or Torch.
    With TensorFlow, a ``tf.Variable`` inside a ``tf.GradientTape`` may be used to
    mark variables as trainable.

    Now we carry out the optimization.
    The minimized cost of the intermediate univariate reconstructions can
    be read out via ``full_output``, including the cost *after* the full Rotosolve step:

    >>> param = init_param
    >>> cost_rotosolve = []
    >>> for step in range(num_steps):
    ...     param, cost, sub_cost = opt.step_and_cost(
    ...         cost_function,
    ...         *param,
    ...         nums_frequency=nums_frequency,
    ...         full_output=True,
    ...         rot_weights=rot_weights,
    ...         crot_weights=crot_weights,
    ...     )
    ...     print(f"Cost before step: {cost}")
    ...     print(f"Minimization substeps: {np.round(sub_cost, 6)}")
    ...     cost_rotosolve.extend(sub_cost)
    Cost before step: 0.04200821039253547
    Minimization substeps: [-0.230905 -0.863336 -0.980072 -0.980072 -1.       -1.       -1.      ]
    Cost before step: -0.9999999990681161
    Minimization substeps: [-1. -1. -1. -1. -1. -1. -1.]
    Cost before step: -0.9999999999999996
    Minimization substeps: [-1. -1. -1. -1. -1. -1. -1.]

    The optimized values for the parameters are now stored in ``param``
    and the optimization behaviour can be assessed by plotting ``cost_rotosolve``,
    which include the substeps of the Rotosolve optimization.
    The ``full_output`` feature is available for both, ``step`` and ``step_and_cost``.

    In general, the frequencies in a QNode will not be integer-valued, requiring us
    to provide the ``RotosolveOptimizer`` not only with the number of frequencies
    but their concrete values. For the example QNode above, this happens if the
    weights are no longer one:

    >>> rot_weights = np.array([0.4, 0.8, 1.2], requires_grad=False)
    >>> crot_weights = np.array([0.5, 1.0, 1.5], requires_grad=False)
    >>> spectrum_fn = qml.fourier.qnode_spectrum(cost_function)
    >>> spectra = spectrum_fn(*param, rot_weights=rot_weights, crot_weights=crot_weights)
    >>> spectra["rot_param"]
    {(0,): [-0.4, 0.0, 0.4], (1,): [-0.8, 0.0, 0.8], (2,): [-1.2, 0.0, 1.2]}
    >>> spectra["crot_param"]
    {(0,): [-0.5, -0.25, 0.0, 0.25, 0.5], (1,): [-1.0, -0.5, 0.0, 0.5, 1.0], (2,): [-1.5, -0.75, 0.0, 0.75, 1.5]}

    We may provide these spectra instead of ``nums_frequency`` to Rotosolve to
    enable the optimization of the QNode at these weights:

    >>> param = init_param
    >>> for step in range(num_steps):
    ...     param, cost, sub_cost = opt.step_and_cost(
    ...         cost_function,
    ...         *param,
    ...         spectra=spectra,
    ...         full_output=True,
    ...         rot_weights = rot_weights,
    ...         crot_weights = crot_weights,
    ...     )
    ...     print(f"Cost before step: {cost}")
    ...     print(f"Minimization substeps: {np.round(sub_cost, 6)}")
    Cost before step: 0.09299359486191039
    Minimization substeps: [-0.268008 -0.713209 -0.24993  -0.871989 -0.907672 -0.907892 -0.940474]
    Cost before step: -0.9404742138557066
    Minimization substeps: [-0.940474 -1.       -1.       -1.       -1.       -1.       -1.      ]
    Cost before step: -1.0
    Minimization substeps: [-1. -1. -1. -1. -1. -1. -1.]

    As we can see, while the optimization got a bit harder and the optimizer takes a bit longer
    to converge than previously, Rotosolve was able to adapt to the more complicated
    dependence on the input arguments and still found the global minimum successfully.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, substep_optimizer="brute", substep_kwargs=None):
        self.substep_kwargs = {} if substep_kwargs is None else substep_kwargs
        if substep_optimizer == "brute":
            self.substep_optimizer = _brute_optimizer
            self.substep_kwargs.setdefault("num_steps", 4)
            self.substep_kwargs.setdefault("Ns", 100)
        elif substep_optimizer == "shgo":
            self.substep_optimizer = _shgo_optimizer
        else:
            self.substep_optimizer = substep_optimizer

    def step_and_cost(
        self,
        objective_fn,
        *args,
        nums_frequency=None,
        spectra=None,
        shifts=None,
        full_output=False,
        **kwargs,
    ):
        r"""Update args with one step of the optimizer and return the corresponding objective
        function value prior to the step. Each step includes multiple substeps, one per
        parameter.

        Args:
            objective_fn (function): the objective function for optimization. It should take a
                sequence of the values ``*args`` and a list of the gates ``generators`` as inputs,
                and return a single value.
            *args (Sequence): variable length sequence containing the initial values of the
                variables to be optimized over or a single float with the initial value.
            nums_frequency (dict[dict]): The number of frequencies in the ``objective_fn`` per
                parameter. The keys must correspond to argument names of the objective
                function, the values must be dictionaries that map parameter indices (``tuple``)
                in the argument to the number of frequencies with which it enters the objective
                function (``int``).
                The parameter index for a scalar QNode argument is ``()``, for
                one-dimensional array QNode arguments, it takes the form ``(i,)`` for the
                i-th parameter in the argument.
            spectra (dict[dict]): Frequency spectra in the ``objective_fn`` per parameter.
                The formatting is the same as for ``nums_frequency``, but the values
                of the inner dictionaries must be sequences of frequencies
                (``Sequence[float]``).
                For each parameter, ``num_frequency`` take precedence over ``spectra``.
            shifts (dict[dict]): Shift angles for the reconstruction per QNode parameter.
                The keys have to be argument names of ``qnode`` and the inner dictionaries have to
                be mappings from parameter indices to the respective shift angles to be used for
                that parameter. For :math:`R` non-zero frequencies, there must be :math:`2R+1`
                shifts given. Ignored if ``nums_frequency`` gives a number of frequencies
                for the respective parameter in the QNode argument.
            full_output (bool): whether to return the intermediate minimized energy values from
                the univariate optimization substeps.
            **kwargs : variable length keyword arguments for the objective function.

        Returns:
            list [array] or array: the new variable values :math:`x^{(t+1)}`.
            If a single arg is provided, list [array] is replaced by array.
            float: the objective function output prior to the step.
            list [float]: the intermediate objective values, only returned if
            ``full_output=True``.

        The optimization step consists of multiple substeps.

        For each substep,
        one of the parameters in one of the QNode arguments is singled out, and the
        objective function is considered as univariate function (i.e., function that
        depends on a single scalar) of that parameter.

        If ``nums_frequency`` states that there is only a single frequency, or ``spectra``
        only contains one positive frequency, for a parameter, an analytic formula is
        used to return the minimum of the univariate restriction.

        For multiple frequencies, :func:`.fourier.reconstruct` is used to reconstruct
        the univariate restriction and a numeric minimization is performed instead.
        The latter minimization is performed using the ``substep_optimizer`` passed to
        ``RotosolveOptimizer`` at initialization.

        .. note::

            One of ``nums_frequency`` and ``spectra`` must contain information
            about each parameter that is to be trained with ``RotosolveOptimizer``.
            For each univariate reconstruction, the data in ``nums_frequency`` takes
            precedence over the information in ``spectra``.

        """
        # todo: does this signature call cover all cases?
        sign_fn = (
            objective_fn.func if isinstance(objective_fn, qml.QNode) else objective_fn
        )
        arg_names = list(signature(sign_fn).parameters.keys())
        requires_grad = {
            arg_name: True
            for arg_name, arg in zip(
                arg_names, args
            )  # changed from the original PennyLane code
        }
        nums_frequency = nums_frequency or {}
        spectra = spectra or {}
        _validate_inputs(requires_grad, args, nums_frequency, spectra)

        # we will single out one arg to change at a time
        # the following hold the arguments not getting updated
        before_args = []
        after_args = list(args)

        # Prepare intermediate minimization results cache
        if full_output:
            y_output = []
        # Compute the very first evaluation in order to be able to cache it
        fun_at_zero = objective_fn(*args, **kwargs)
        first_substep_in_step = True

        for arg_idx, (arg, arg_name) in enumerate(zip(args, arg_names)):
            del after_args[0]

            if not requires_grad[arg_name]:
                before_args.append(arg)
                continue
            shape = qml.math.shape(arg)
            indices = np.ndindex(shape) if len(shape) > 0 else [()]
            for par_idx in indices:
                _fun_at_zero = fun_at_zero if first_substep_in_step else None
                # Set a single parameter in a single argument to be reconstructed
                num_freq = nums_frequency.get(arg_name, {}).get(par_idx, None)
                spectrum = spectra.get(arg_name, {}).get(par_idx, None)
                if spectrum is not None:
                    spectrum = np.array(spectrum)

                if (
                    num_freq == 1
                    or (spectrum is not None and len(spectrum[spectrum > 0])) == 1
                ):
                    _args = before_args + [arg] + after_args
                    univariate = _restrict_to_univariate(
                        objective_fn, arg_idx, par_idx, _args, kwargs
                    )
                    freq = 1.0 if num_freq is not None else spectrum[spectrum > 0][0]
                    x_min, y_min = self.min_analytic(univariate, freq, _fun_at_zero)
                    arg = qml.math.scatter_element_add(arg, par_idx, x_min)

                else:
                    ids = {arg_name: (par_idx,)}
                    _nums_frequency = (
                        {arg_name: {par_idx: num_freq}}
                        if num_freq is not None
                        else None
                    )
                    _spectra = (
                        {arg_name: {par_idx: spectrum}}
                        if spectrum is not None
                        else None
                    )

                    # Set up the reconstruction function
                    recon_fn = qml.fourier.reconstruct(
                        objective_fn, ids, _nums_frequency, _spectra, shifts
                    )
                    # Perform the reconstruction
                    recon = recon_fn(
                        *before_args, arg, *after_args, f0=_fun_at_zero, **kwargs
                    )[arg_name][par_idx]
                    if spectrum is None:
                        spectrum = list(range(num_freq + 1))
                    x_min, y_min = self._min_numeric(recon, spectrum)

                    # Update the currently treated argument
                    arg = qml.math.scatter_element_add(
                        arg, par_idx, x_min - arg[par_idx]
                    )
                first_substep_in_step = False

                if full_output:
                    y_output.append(y_min)

            # updating before_args for next argument
            before_args.append(arg)

        # All arguments have been updated and/or passed to before_args
        args = before_args
        # unwrap arguments if only one, backward compatible and cleaner
        if len(args) == 1:
            args = args[0]

        if full_output:
            return args, fun_at_zero, y_output

        return args, fun_at_zero

    def step(
        self,
        objective_fn,
        *args,
        nums_frequency=None,
        spectra=None,
        shifts=None,
        full_output=False,
        **kwargs,
    ):
        r"""Update args with one step of the optimizer. Each step includes
        multiple substeps, one per parameter.

        Args:
            objective_fn (function): the objective function for optimization. It should take a
                sequence of the values ``*args`` and a list of the gates ``generators`` as inputs,
                and return a single value.
            *args (Sequence): variable length sequence containing the initial values of the
                variables to be optimized over or a single float with the initial value.
            nums_frequency (dict[dict]): The number of frequencies in the ``objective_fn`` per
                parameter. The keys must correspond to argument names of the objective
                function, the values must be dictionaries that map parameter indices (``tuple``)
                in the argument to the number of frequencies with which it enters the objective
                function (``int``).
                The parameter index for a scalar QNode argument is ``()``, for
                one-dimensional array QNode arguments, it takes the form ``(i,)`` for the
                i-th parameter in the argument.
            spectra (dict[dict]): Frequency spectra in the ``objective_fn`` per parameter.
                The formatting is the same as for ``nums_frequency``, but the values
                of the inner dictionaries must be sequences of frequencies
                (``Sequence[float]``).
                For each parameter, ``num_frequency`` take precedence over ``spectra``.
            shifts (dict[dict]): Shift angles for the reconstruction per QNode parameter.
                The keys have to be argument names of ``qnode`` and the inner dictionaries have to
                be mappings from parameter indices to the respective shift angles to be used for
                that parameter. For :math:`R` non-zero frequencies, there must be :math:`2R+1`
                shifts given. Ignored if ``nums_frequency`` gives a number of frequencies
                for the respective parameter in the QNode argument.
            full_output (bool): whether to return the intermediate minimized energy values from
                the univariate optimization substeps.
            **kwargs : variable length keyword arguments for the objective function.

        Returns:
            list [array] or array: the new variable values :math:`x^{(t+1)}`.
                If a single arg is provided, list [array] is replaced by array.
            list [float]: the intermediate objective values, only returned if
                ``full_output=True``.

        The optimization step consists of multiple substeps.

        For each substep,
        one of the parameters in one of the QNode arguments is singled out, and the
        objective function is considered as univariate function (i.e., function that
        depends on a single scalar) of that parameter.

        If ``nums_frequency`` states that there is only a single frequency, or ``spectra``
        only contains one positive frequency, for a parameter, an analytic formula is
        used to return the minimum of the univariate restriction.

        For multiple frequencies, :func:`.fourier.reconstruct` is used to reconstruct
        the univariate restriction and a numeric minimization is performed instead.
        The latter minimization is performed using the ``substep_optimizer`` passed to
        ``RotosolveOptimizer`` at initialization.

        .. note::

            One of ``nums_frequency`` and ``spectra`` must contain information
            about each parameter that is to be trained with ``RotosolveOptimizer``.
            For each univariate reconstruction, the data in ``nums_frequency`` takes
            precedence over the information in ``spectra``.

        """
        x_new, _, *y_output = self.step_and_cost(
            objective_fn,
            *args,
            nums_frequency=nums_frequency,
            spectra=spectra,
            shifts=shifts,
            full_output=full_output,
            **kwargs,
        )
        if full_output:
            # For full_output=True, y_output was wrapped in an outer list due
            # to the dynamic unpacking
            return x_new, y_output[0]

        return x_new

    def _min_numeric(self, objective_fn, spectrum):
        r"""Numerically minimize a trigonometric function that depends on a
        single parameter. Uses potentially large numbers of function evaluations,
        depending on the used substep_optimizer. The optimization method and
        options are stored in ``RotosolveOptimizer.substep_optimizer``
        and ``RotosolveOptimizer.substep_kwargs``.

        Args:
            objective_fn (callable): Trigonometric function to minimize

        Returns:
            float: Position of the minimum of ``objective_fn``
            float: Value of the minimum of ``objective_fn``

        The returned position is guaranteed to lie within :math:`(-\pi, \pi]`.
        """
        opt_kwargs = self.substep_kwargs.copy()
        if "bounds" not in self.substep_kwargs:
            spectrum = qml.math.array(spectrum)
            half_width = np.pi / qml.math.min(spectrum[spectrum > 0])
            opt_kwargs["bounds"] = ((-half_width, half_width),)

        x_min, y_min = self.substep_optimizer(objective_fn, **opt_kwargs)
        if y_min is None:
            y_min = objective_fn(x_min)

        return x_min, y_min

    @staticmethod
    def min_analytic(objective_fn, freq, f0):
        r"""Analytically minimize a trigonometric function that depends on a
        single parameter and has a single frequency. Uses two or
        three function evaluations.

        Args:
            objective_fn (callable): Trigonometric function to minimize
            freq (float): Frequency :math:`f` in the ``objective_fn``
            f0 (float): Value of the ``objective_fn`` at zero. Reduces the
                number of calls to the function from three to two if given.

        Returns:
            float: Position of the minimum of ``objective_fn``
            float: Value of the minimum of ``objective_fn``

        The closed form expression used here was derived in
        `Vidal & Theis (2018) <https://arxiv.org/abs/1812.06323>`__ ,
        `Parrish et al (2019) <https://arxiv.org/abs/1904.03206>`__ and
        `Ostaszewski et al (2021) <https://doi.org/10.22331/q-2021-01-28-391>`__.
        We use the notation of Appendix A of the last of these references,
        although we allow for an arbitrary frequency instead of restricting
        to :math:`f=1`.
        The returned position is guaranteed to lie within :math:`(-\pi/f, \pi/f]`.

        The used formula for the minimization of the :math:`d-\text{th}`
        parameter then reads

        .. math::

            \theta^*_d &= \underset{\theta_d}{\text{argmin}}\left<H\right>_{\theta_d}\\
                  &= -\frac{\pi}{2f} - \frac{1}{f}\text{arctan2}\left(2\left<H\right>_{\theta_d=0}
                  - \left<H\right>_{\theta_d=\pi/(2f)} - \left<H\right>_{\theta_d=-\pi/(2f)},
                  \left<H\right>_{\theta_d=\pi/(2f)} - \left<H\right>_{\theta_d=-\pi/(2f)}\right),

        """
        if f0 is None:
            f0 = objective_fn(0.0)
        shift = 0.5 * np.pi / freq
        fp = objective_fn(shift)
        fm = objective_fn(-shift)
        C = 0.5 * (fp + fm)
        B = np.arctan2(2 * f0 - fp - fm, fp - fm)
        x_min = -shift - B / freq
        A = np.sqrt((f0 - C) ** 2 + 0.25 * (fp - fm) ** 2)
        y_min = -A + C

        if x_min <= -2 * shift:
            x_min = x_min + 4 * shift

        return x_min, y_min



import time
import numpy as np

from .rqaoa_workflow_properties import RqaoaParameters
from ..baseworkflow import Workflow, check_compiled
from ..qaoa import QAOA
from ..workflow_properties import CircuitProperties
from ...backends.devices_core import DeviceLocal, DeviceBase
from ...problems import QUBO
from ...utilities import (
    ground_state_hamiltonian,
    exp_val_hamiltonian_termwise,
    generate_timestamp,
)
from ...backends.qaoa_analytical_sim import QAOABackendAnalyticalSimulator
from . import rqaoa_utils
from .rqaoa_result import RQAOAResult


class RQAOA(Workflow):
    """
    A class implementing a RQAOA workflow end to end.

    It's basic usage consists of
    1. Initialization
    2. Compilation
    3. Optimization

    .. note::
        The attributes of the RQAOA class should be initialized using the set methods of QAOA.
        For example, to set the qaoa circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
    device: `DeviceBase`
        Device to be used by the optimizer
    backend_properties: `BackendProperties`
        The backend properties of the RQAOA workflow. These properties will be used to
        run QAOA at each RQAOA step. Use to set the backend properties such as the
        number of shots and the cvar values. For a complete list of its parameters and
        usage please see the method set_backend_properties
    classical_optimizer: `ClassicalOptimizer`
        The classical optimiser properties of the RQAOA workflow.
        Use to set the classical optimiser needed for the classical optimisation part of the QAOA routine.
        For a complete list of its parameters and usage please see the method set_classical_optimizer
    local_simulators: `list[str]`
        A list containing the available local simulators
    cloud_provider: `list[str]`
        A list containing the available cloud providers
    compiled: `Bool`
        A boolean flag to check whether the optimizer object has been correctly compiled at least once
    circuit_properties: `CircuitProperties`
        The circuit properties of the RQAOA workflow. These properties will be used to run QAOA at each RQAOA step.
        Use to set depth `p`, choice of parametrisation, parameter initialisation strategies, mixer hamiltonians.
        For a complete list of its parameters and usage please see the method set_circuit_properties
    rqaoa_parameters: `RqaoaParameters`
        Set of parameters containing all the relevant information for the recursive procedure of RQAOA.
    results: `RQAOAResult`
        The results of the RQAOA optimization.
        Dictionary containing all the information about the RQAOA run: the
        solution states and energies (key: 'solution'), the output of the classical
        solver (key: 'classical_output'), the elimination rules for each step
        (key: 'elimination_rules'), the number of eliminations at each step (key: 'schedule'),
        total number of steps (key: 'number_steps'), the intermediate QUBO problems and the
        intermediate QAOA objects that have been optimized in each RQAOA step (key: 'intermediate_problems').
        This object (`RQAOAResult`) is a dictionary with some custom methods as
        RQAOAResult.get_hamiltonian_step(i) which get the hamiltonian of reduced problem of the i-th step.
        To see the full list of methods please see the RQAOAResult class.

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> r = RQAOA()
    >>> r.compile(QUBO)
    >>> r.optimize()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    Standard/custom (default) type:

    >>> r = RQAOA()
    >>> r.set_circuit_properties(
            p=10,
            param_type='extended',
            init_type='ramp',
            mixer_hamiltonian='x'
        )
    >>> r.set_device_properties(
            device_location='qcs',
            device_name='Aspen-11',
            cloud_credentials={
                'name' : "Aspen11",
                'as_qvm':True,
                'execution_timeout' : 10,
                'compiler_timeout':10
            }
        )
    >>> r.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> r.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> r.set_rqaoa_parameters(n_cutoff = 5, steps=[1,2,3,4,5])
    >>> r.compile(qubo_problem)
    >>> r.optimize()

    Ada-RQAOA:

    >>> r_adaptive = RQAOA()
    >>> r.set_circuit_properties(
            p=10,
            param_type='extended',
            init_type='ramp',
            mixer_hamiltonian='x'
        )
    >>> r.set_device_properties(
            device_location='qcs',
            device_name='Aspen-11',
            cloud_credentials={
                'name' : "Aspen11",
                'as_qvm':True,
                'execution_timeout' : 10,
                'compiler_timeout':10
            }
        )
    >>> r_adaptive.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> r_adaptive.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> r_adaptive.set_rqaoa_parameters(rqaoa_type = 'adaptive', n_cutoff = 5, n_max=5)
    >>> r_adaptive.compile(qubo_problem)
    >>> r_adaptive.optimize()
    """

    results_class = RQAOAResult

    def __init__(self, device: DeviceBase = DeviceLocal("vectorized")):
        """
        Initialize the RQAOA class.

        Parameters
        ----------
            device: `DeviceBase`
                Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        super().__init__(device)  # use the parent class to initialize
        self.circuit_properties = CircuitProperties()
        self.rqaoa_parameters = RqaoaParameters()

        # change algorithm name to rqaoa
        self.header["algorithm"] = "rqaoa"

    @check_compiled
    def set_circuit_properties(self, **kwargs):
        """
        Specify the circuit properties to construct the QAOA circuits

        Parameters
        ----------
            qubit_register: `list`
                Select the desired qubits to run the QAOA program. Meant to be used as a qubit
                selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
            p: `int`
                Depth `p` of the QAOA circuit
            q: `int`
                Analogue of `p` of the QAOA circuit in the Fourier parameterisation
            param_type: `str`
                Choose the QAOA circuit parameterisation. Currently supported parameterisations include:
                `'standard'`: Standard QAOA parameterisation
                `'standard_w_bias'`: Standard QAOA parameterisation with a separate parameter
                for single-qubit terms.
                `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
                `'fourier'`: Fourier circuit parameterisation
                `'fourier_extended'`: Fourier circuit parameterisation with individual parameter
                for each qubit and term in Hamiltonian.
                `'fourier_w_bias'`: Fourier circuit parameterisation with a separate parameter
                for single-qubit terms
            init_type: `str`
                Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
                `'rand'`: Randomly initialise circuit parameters
                `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit parameters
                (inspired from Quantum Annealing)
                `'custom'`: User specified initial circuit parameters
            mixer_hamiltonian: `str`
                Parameterisation of the mixer hamiltonian:
                `'x'`: Randomly initialise circuit parameters
                `'xy'`: Linear ramp from Hamiltonian initialisation of circuit
            mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
                The connectivity of the qubits in the mixer Hamiltonian. Use only if `mixer_hamiltonian = xy`.
                The user can specify the connectivity as a list of lists, a list of tuples,
                or a string chosen from ['full', 'chain', 'star'].
            mixer_coeffs: `list`
                The coefficients of the mixer Hamiltonian. By default all set to -1
            annealing_time: `float`
                Total time to run the QAOA program in the Annealing parameterisation (digitised annealing)
            linear_ramp_time: `float`
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            variational_params_dict: `dict`
                Dictionary object specifying the initial value of each circuit parameter for
                the chosen parameterisation, if the `init_type` is selected as `'custom'`.
                For example, for standard params set {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
        """

        for key in kwargs.keys():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError(
                    f"Specified argument {key} is not supported by the circuit"
                )

        self.circuit_properties = CircuitProperties(**kwargs)

        return None

    @check_compiled
    def set_rqaoa_parameters(self, **kwargs):
        """
        Specify the parameters to run a desired RQAOA program.

        Parameters
        ----------
        rqaoa_type: `int`
            String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
            'adaptive'. Defaults to 'custom'.
        n_max: `int`
            Maximum number of eliminations allowed at each step when using the adaptive method.
        steps: `Union[list,int]`
            Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins eliminated
            at each step. If a list is passed, the algorithm will follow the list to select how many spins to eliminate
            at each step. Note that the list needs enough elements to specify eliminations from the initial number of qubits
            up to the cutoff value. If the list contains more, the algorithm will follow instructions until the cutoff value
            is reached.
        n_cutoff: `int`
            Cutoff value at which the RQAOA algorithm obtains the solution classically.
        original_hamiltonian: `Hamiltonian`
            Hamiltonian encoding the original problem fed into the RQAOA algorithm.
        counter: `int`
            Variable to count the step in the schedule. If counter = 3 the next step is schedule[3].
            Default is 0, but can be changed to start in the position of the schedule that one wants.
        """

        for key in kwargs.keys():
            if hasattr(self.rqaoa_parameters, key):
                pass
            else:
                raise ValueError(f"Specified argument {key} is not supported by RQAOA")

        self.rqaoa_parameters = RqaoaParameters(**kwargs)

        return None

    def compile(self, problem: QUBO = None, verbose: bool = False):
        """
        Create a QAOA object and initialize it with the circuit properties, device, classical optimizer and
        backend properties specified by the user.
        This QAOA object will be used to run QAOA changing the problem to sove at each RQAOA step.
        Here, the QAOA is compiled passing the problem statement, so to check that the compliation of
        QAOA is correct. See the QAOA class.

        .. note::
            Compilation is necessary because it is the moment where the problem statement and
            the QAOA instructions are used to build the actual QAOA circuit.

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by RQAOA
        verbose: bool
            !NotYetImplemented! Set true to have a summary of QAOA first step to displayed after compilation
        """

        # we compile the method of the parent class to genereate the id
        # and check the problem is a QUBO object and save it
        super().compile(problem=problem)

        # if type is custom and steps is an int, set steps correctly
        if (
            self.rqaoa_parameters.rqaoa_type == "custom"
            and self.rqaoa_parameters.n_cutoff <= problem.n
        ):
            n_cutoff = self.rqaoa_parameters.n_cutoff
            n_qubits = problem.n
            counter = self.rqaoa_parameters.counter

            # If schedule for custom RQAOA is not given, we create a schedule such that
            # n = self.rqaoa_parameters.steps spins is eliminated at a time
            if type(self.rqaoa_parameters.steps) is int:
                self.rqaoa_parameters.steps = [self.rqaoa_parameters.steps] * (
                    n_qubits - n_cutoff
                )

            # In case a schedule is given, ensure there are enough steps in the schedule
            assert np.abs(n_qubits - n_cutoff - counter) <= sum(
                self.rqaoa_parameters.steps
            ), f"Schedule is incomplete, add {np.abs(n_qubits - n_cutoff - counter) - sum(self.rqaoa_parameters.steps)} more eliminations"

        # Create the qaoa object with the properties
        self.__q = QAOA(self.device)
        self.__q.circuit_properties = self.circuit_properties
        self.__q.backend_properties = self.backend_properties
        self.__q.classical_optimizer = self.classical_optimizer
        self.__q.exp_tags = self.exp_tags

        # set the header of the qaoa object to be the same as the header of the rqaoa object
        self.__q.header = self.header.copy()
        self.__q.header[
            "algorithm"
        ] = "qaoa"  # change the algorithm name to qaoa, since this is a qaoa object

        # compile qaoa object
        self.__q.compile(problem, verbose=verbose)

        # set compiled boolean to true
        self.compiled = True

        return

    def optimize(
        self, dump: bool = False, dump_options: dict = {}, verbose: bool = False
    ):
        """
        Performs optimization using RQAOA with the `custom` method or the `adaptive` method.
        The elimination RQAOA loop will occur until the number of qubits is equal to the number of qubits specified in `n_cutoff`.
        In each loop, the QAOA will be run, then the eliminations will be computed, a new problem will be redefined
        and the QAOA will be recompiled with the new problem.
        Once the loop is complete, the final problem will be solved classically and the final solution will be reconstructed.
        Results will be stored in the `results` attribute.

        Parameters
        ----------
        dump: `bool`
            If true, the object will be dumped to a file. And at the end of each step,
            the qaoa object will be dumped to a file.
            Default is False.
        dump_options: `dict`
            Dictionary containing the options for the dump. To see the options,
            see the `dump` method of the `QAOA` or `RQAOA` class.
            Default is empty.
        verbose: `bool`
            TODO
        """

        # check if the object has been compiled (or already optimized)
        assert (
            self.compiled
        ), "RQAOA object has not been compiled. Please run the compile method first."

        # lists to append the eliminations, the problems, the qaoa results objects,
        # the correlation matrix, the expectation values z and a dictionary for the atomic ids
        elimination_tracker = []
        q_results_steps = []
        problem_steps = []
        exp_vals_z_steps = []
        corr_matrix_steps = []
        atomic_id_steps = {}

        # get variables
        problem = self.problem
        problem_metadata = self.problem.metadata
        n_cutoff = self.rqaoa_parameters.n_cutoff
        n_qubits = problem.n
        counter = self.rqaoa_parameters.counter

        # get the qaoa object
        q = self.__q

        # create a different max_terms function for each type
        if self.rqaoa_parameters.rqaoa_type == "adaptive":
            f_max_terms = rqaoa_utils.ada_max_terms
        else:
            f_max_terms = rqaoa_utils.max_terms

        # timestamp for the start of the optimization
        self.header["execution_time_start"] = generate_timestamp()

        # flag, set to true if the problem vanishes due to elimination before reaching cutoff
        total_elimination = False

        # If above cutoff, loop quantumly, else classically
        while n_qubits > n_cutoff:
            # put a tag to the qaoa object to know which step it is
            q.set_exp_tags({"rqaoa_counter": counter})

            # Run QAOA
            q.optimize()

            # save the results if dump is true
            if dump:
                q.dump(**dump_options)

            # Obtain statistical results
            exp_vals_z, corr_matrix = self.__exp_val_hamiltonian_termwise(q)
            # Retrieve highest expectation values according to adaptive method or schedule in custom method
            max_terms_and_stats = f_max_terms(
                exp_vals_z, corr_matrix, self.__n_step(n_qubits, n_cutoff, counter)
            )
            # Generate spin map
            spin_map = rqaoa_utils.spin_mapping(problem, max_terms_and_stats)
            # Eliminate spins and redefine problem
            new_problem, spin_map = rqaoa_utils.redefine_problem(problem, spin_map)

            # In case eliminations cancel out the whole graph, break the loop before reaching the predefined cutoff.
            if new_problem == problem:
                total_elimination = True
                break

            # Extract final set of eliminations with correct dependencies and update tracker
            eliminations = [
                {"singlet": (spin,), "bias": spin_map[spin][0]}
                if spin_map[spin][1] is None
                else {
                    "pair": (spin_map[spin][1], spin),
                    "correlation": spin_map[spin][0],
                }
                for spin in sorted(spin_map.keys())
                if spin != spin_map[spin][1]
            ]
            elimination_tracker.append(eliminations)
            # add the metadata to the problem
            new_problem.metadata = problem_metadata

            # Save qaoa object, correlation matrix and expectation values z
            q_results_steps.append(q.result)
            corr_matrix_steps.append(corr_matrix)
            exp_vals_z_steps.append(exp_vals_z)
            problem_steps.append(problem)
            atomic_id_steps[counter] = q.header["atomic_id"]

            # Extract new number of qubits
            n_qubits = new_problem.n

            # problem is updated
            problem = new_problem

            # Compile qaoa with the problem
            q.compile(problem, verbose=False)

            # Add one step to the counter
            counter += 1
            # TODO: do rqaoa dumps here if dump is true, so the user can still get the results in case the loop is interrupted.

        if total_elimination:
            # Solve the smallest non-vanishing problem by fixing spins arbitrarily or according to the correlations
            cl_energy, cl_ground_states = rqaoa_utils.solution_for_vanishing_instances(
                problem.hamiltonian, spin_map
            )
        else:
            # Solve the new problem classically
            cl_energy, cl_ground_states = ground_state_hamiltonian(problem.hamiltonian)

        # Retrieve full solutions including eliminated spins and their energies
        full_solutions = rqaoa_utils.final_solution(
            elimination_tracker, cl_ground_states, self.problem.hamiltonian
        )

        # timestamp for the end of the optimization
        self.header["execution_time_end"] = generate_timestamp()

        # Compute description dictionary containing all the information
        self.result = self.results_class()
        self.result["solution"] = full_solutions
        self.result["classical_output"] = {
            "minimum_energy": cl_energy,
            "optimal_states": cl_ground_states,
        }
        self.result["elimination_rules"] = elimination_tracker
        self.result["schedule"] = [
            len(eliminations) for eliminations in elimination_tracker
        ]
        self.result["number_steps"] = counter - self.rqaoa_parameters.counter
        self.result["intermediate_steps"] = [
            {
                "counter": counter,
                "problem": problem,
                "qaoa_results": q_results,
                "exp_vals_z": exp_vals_z,
                "corr_matrix": corr_matrix,
            }
            for counter, problem, q_results, exp_vals_z, corr_matrix in zip(
                range(self.rqaoa_parameters.counter, counter),
                problem_steps,
                q_results_steps,
                exp_vals_z_steps,
                corr_matrix_steps,
            )
        ]
        self.result["atomic_ids"] = atomic_id_steps

        # set compiled to false
        self.compiled = False

        # dump the object if dump is true
        if dump:
            self.dump(
                **{**dump_options, **{"options": {"intermediate_measurements": False}}}
            )

        if verbose:
            print(f"RQAOA optimization completed.")

        return

    def __exp_val_hamiltonian_termwise(self, q: QAOA):
        """
        Private method to call the exp_val_hamiltonian_termwise function taking the data from
        the QAOA object _q.
        It eturns what the exp_val_hamiltonian_termwise function returns.
        """

        variational_params = q.variate_params
        qaoa_backend = q.backend
        cost_hamiltonian = q.cost_hamil
        mixer_type = q.circuit_properties.mixer_hamiltonian
        p = q.circuit_properties.p
        qaoa_optimized_angles = q.result.optimized["angles"]
        qaoa_optimized_counts = q.result.get_counts(
            q.result.optimized["measurement_outcomes"]
        )
        analytical = isinstance(qaoa_backend, QAOABackendAnalyticalSimulator)

        return exp_val_hamiltonian_termwise(
            cost_hamiltonian,
            mixer_type,
            p,
            qaoa_optimized_angles,
            qaoa_optimized_counts,
            analytical=analytical,
        )

    def __n_step(self, n_qubits, n_cutoff, counter):
        """
        Private method that returns the n_max value in case of adaptive or the number of eliminations according
        to the schedule and the counter in case of custom method.
        """

        if self.rqaoa_parameters.rqaoa_type == "adaptive":
            # Number of spins to eliminate according the schedule
            n = self.rqaoa_parameters.n_max
        else:
            # max Number of spins to eliminate
            n = self.rqaoa_parameters.steps[counter]

        # If the step eliminates more spins than available, reduce step to match cutoff
        return (n_qubits - n_cutoff) if (n_qubits - n_cutoff) < n else n

    def _serializable_dict(
        self, complex_to_string: bool = False, intermediate_measurements: bool = True
    ):
        """
        Returns all values and attributes of the object that we want to
        return in `asdict` and `dump(s)` methods in a dictionary.

        Parameters
        ----------
        complex_to_string: bool
            If True, complex numbers are converted to strings.
            If False, complex numbers are converted to lists of real and imaginary parts.

        Returns
        -------
        serializable_dict: dict
            Dictionary containing all the values and attributes of the object
            that we want to return in `asdict` and `dump(s)` methods.
        intermediate_measurements: bool
            If True, intermediate measurements are included in the dump. If False,
            intermediate measurements are not included in the dump.
            Default is True.
        """
        # we call the _serializable_dict method of the parent class,
        # specifying the keys to delete from the results dictionary
        serializable_dict = super()._serializable_dict(
            complex_to_string, intermediate_measurements
        )

        # we add the keys of the RQAOA object that we want to return
        serializable_dict["data"]["input_parameters"]["circuit_properties"] = dict(
            self.circuit_properties
        )
        serializable_dict["data"]["input_parameters"]["rqaoa_parameters"] = dict(
            self.rqaoa_parameters
        )

        # include parameters in the header metadata
        serializable_dict["header"]["metadata"]["param_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["param_type"]
        serializable_dict["header"]["metadata"]["init_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["init_type"]
        serializable_dict["header"]["metadata"]["p"] = serializable_dict["data"][
            "input_parameters"
        ]["circuit_properties"]["p"]

        if (
            serializable_dict["data"]["input_parameters"]["circuit_properties"]["q"]
            is not None
        ):
            serializable_dict["header"]["metadata"]["q"] = serializable_dict["data"][
                "input_parameters"
            ]["circuit_properties"]["q"]

        serializable_dict["header"]["metadata"]["rqaoa_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["rqaoa_parameters"]["rqaoa_type"]
        serializable_dict["header"]["metadata"]["rqaoa_n_max"] = serializable_dict[
            "data"
        ]["input_parameters"]["rqaoa_parameters"]["n_max"]
        serializable_dict["header"]["metadata"]["rqaoa_n_cutoff"] = serializable_dict[
            "data"
        ]["input_parameters"]["rqaoa_parameters"]["n_cutoff"]

        return serializable_dict



from __future__ import annotations
from abc import ABC, abstractproperty
from typing import List, Union, Tuple, Any, Callable, Iterable, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ...backends.basedevice import DeviceBase
import numpy as np
from enum import Enum
import copy

from .operators import Hamiltonian
from .hamiltonianmapper import HamiltonianMapper
from .gatemap import RotationGateMap, SWAPGateMap, GateMap
from .gatemaplabel import GateMapType




def _is_iterable_empty(in_iterable):
    if isinstance(in_iterable, Iterable):  # Is Iterable
        return all(map(_is_iterable_empty, in_iterable))
    return False  # Not an Iterable


class shapedArray:
    """Decorator-Descriptor for arrays that have a fixed shape.

    This is used to facilitate automatic input checking for all the different
    internal parameters. Each instance of this class can be removed without
    replacement and the code should still work, provided the user provides
    only correct angles to below parameter classes

    Parameters
    ----------
    shape: `Callable[[Any], Tuple]`
        Returns the shape for self.values

    Example
    -------
    With this descriptor, the two following are equivalent:

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)
                self._my_attribute = None

            @property
            def my_attribute(self):
                return _my_attribute

            @my_attribute.setter
            def my_attribute(self):
                try:
                    self._my_attribute = np.reshape(values, self.shape)
                except ValueError:
                    raise ValueError("my_attribute must have shape "
                                    f"{self.shape}")


    can be simplified to

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)

            @shapedArray
            def my_attribute(self):
                return self.shape
    """

    def __init__(self, shape: Callable[[Any], Tuple]):
        """The constructor. See class documentation"""
        self.name = shape.__name__
        self.shape = shape

    def __set__(self, obj, values):
        """The setter with input checking."""
        try:
            # also round to 12 decimal places to avoid floating point errors
            setattr(
                obj, f"__{self.name}", np.round(np.reshape(values, self.shape(obj)), 12)
            )
        except ValueError:
            raise ValueError(f"{self.name} must have shape {self.shape(obj)}")

    def __get__(self, obj, objtype):
        """The getter."""
        return getattr(obj, f"__{self.name}")


class AnsatzDescriptor(ABC):
    """
    Parameters class to construct a specific quantum ansatz to attack
    a problem

    Parameters
    ----------
    algorithm: `str`
        The algorithm corresponding to the ansatz
    Attributes
    ----------
    algorithm: `str`
    """

    def __init__(self, algorithm: str):
        self.algorithm = algorithm

    @abstractproperty
    def n_qubits(self) -> int:
        pass


class QAOADescriptor(AnsatzDescriptor):

    """
    Create the problem attributes consisting of the Hamiltonian, QAOA 'p'
    value and other specific parameters.

    Attributes
    ----------
    cost_hamiltonian: `Hamiltonian`

    qureg: `List[int]`

    cost_block_coeffs: `List[float]`

    cost_single_qubit_coeffs: `List[float]`

    cost_qubits_singles: `List[str]`

    cost_pair_qubit_coeffs: `List[float]`

    cost_qubits_pairs: `List[str]`

    mixer_block_coeffs: `List[float]`

    cost_blocks: `List[RotationGateMap]`

    mixer_blocks: `List[RotationGateMap]`

    Properties
    ----------
    n_qubits: `int`

    abstract_circuit: `List[RotationGateMap]`
    """

    def __init__(
        self,
        cost_hamiltonian: Hamiltonian,
        mixer_block: Union[List[RotationGateMap], Hamiltonian],
        p: int,
        mixer_coeffs: List[float] = [],
        routing_function: Optional[Callable] = None,
        device: Optional["DeviceBase"] = None,
    ):
        """
        Parameters
        ----------
        cost_hamiltonian: `Hamiltonian`
            The cost hamiltonian of the problem the user is trying to solve.
        mixer_block: Union[List[RotationGateMap], Hamiltonian]
            The mixer hamiltonian or a list of initialised RotationGateMap objects
            that defines the gates to be used within the "mixer part" of the circuit.
        p: `int`
            Number of QAOA layers; defaults to 1 if not specified
        mixer_coeffs: `List[float]`
            A list containing coefficients for each mixer GateMap. The order of the
            coefficients should follow the order of the GateMaps provided in the relevant gate block.
            This input isnt required if the input mixer block is of type Hamiltonian.
        routing_function Optional[Callable]
            A callable function running the routing algorithm on the problem
        device: DeviceBase
            The device on which to run the Quantum Circuit
        """

        super().__init__(algorithm="QAOA")

        self.p = p
        self.cost_block_coeffs = cost_hamiltonian.coeffs

        try:
            self.mixer_block_coeffs = mixer_block.coeffs
        except AttributeError:
            self.mixer_block_coeffs = mixer_coeffs

        # Needed in the BaseBackend to compute exact_solution, cost_funtion method
        # and bitstring_energy
        self.cost_hamiltonian = cost_hamiltonian
        self.cost_block = self.block_setter(cost_hamiltonian, GateMapType.COST)
        (
            self.cost_single_qubit_coeffs,
            self.cost_pair_qubit_coeffs,
            self.cost_qubits_singles,
            self.cost_qubits_pairs,
        ) = self._assign_coefficients(self.cost_block, self.cost_block_coeffs)

        # route the cost block and append SWAP gates
        if isinstance(routing_function, Callable):
            try:
                (
                    self.cost_block,
                    self.initial_mapping,
                    self.final_mapping,
                ) = self.route_gates_list(self.cost_block, device, routing_function)
                self.routed = True
            except TypeError:
                raise TypeError(
                    "The specified function can has a set signature that accepts"
                    " device, problem, and initial_mapping"
                )
            except Exception as e:
                raise e
        elif routing_function == None:
            self.routed = False
        else:
            raise ValueError(
                f"Routing function can only be a Callable not {type(routing_function)}"
            )

        self.mixer_block = self.block_setter(mixer_block, GateMapType.MIXER)
        (
            self.mixer_single_qubit_coeffs,
            self.mixer_pair_qubit_coeffs,
            self.mixer_qubits_singles,
            self.mixer_qubits_pairs,
        ) = self._assign_coefficients(self.mixer_block, self.mixer_block_coeffs)

        self.mixer_blocks = HamiltonianMapper.repeat_gate_maps(self.mixer_block, self.p)
        self.cost_blocks = HamiltonianMapper.repeat_gate_maps(self.cost_block, self.p)
        self.qureg = list(range(self.n_qubits))

    @property
    def n_qubits(self) -> int:
        if self.routed == True:
            return len(self.final_mapping)
        else:
            return self.cost_hamiltonian.n_qubits

    def __repr__(self):
        """Return an overview over the parameters and hyperparameters

        Todo
        ----
        Split this into ``__repr__`` and ``__str__`` with a more verbose
        output in ``__repr__``.
        """

        string = "Circuit Parameters:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tregister: " + str(self.qureg) + "\n" + "\n"

        string += "Cost Hamiltonian:\n"
        string += "\tcost_qubits_singles: " + str(self.cost_qubits_singles) + "\n"
        string += (
            "\tcost_single_qubit_coeffs: " + str(self.cost_single_qubit_coeffs) + "\n"
        )
        string += "\tcost_qubits_pairs: " + str(self.cost_qubits_pairs) + "\n"
        string += (
            "\tcost_pair_qubit_coeffs: "
            + str(self.cost_pair_qubit_coeffs)
            + "\n"
            + "\n"
        )

        string += "Mixer Hamiltonian:\n"
        string += "\tmixer_qubits_singles: " + str(self.mixer_qubits_singles) + "\n"
        string += (
            "\tmixer_single_qubit_coeffs: " + str(self.mixer_single_qubit_coeffs) + "\n"
        )
        string += "\tmixer_qubits_pairs: " + str(self.mixer_qubits_pairs) + "\n"
        string += (
            "\tmixer_pair_qubit_coeffs: " + str(self.mixer_pair_qubit_coeffs) + "\n"
        )

        return string

    def _assign_coefficients(
        self, input_block: List[RotationGateMap], input_coeffs: List[float]
    ) -> None:
        """
        Splits the coefficients and gatemaps into qubit singles and qubit pairs.
        """

        single_qubit_coeffs = []
        pair_qubit_coeffs = []
        qubit_singles = []
        qubit_pairs = []

        if len(input_block) != len(input_coeffs):
            raise ValueError(
                "The number of terms/gatemaps must match the number of coefficients provided."
            )
        for each_gatemap, each_coeff in zip(input_block, input_coeffs):
            if each_gatemap.gate_label.n_qubits == 1:
                single_qubit_coeffs.append(each_coeff)
                # Giving a string name to each gatemap (?)
                qubit_singles.append(type(each_gatemap).__name__)
            elif each_gatemap.gate_label.n_qubits == 2:
                pair_qubit_coeffs.append(each_coeff)
                qubit_pairs.append(type(each_gatemap).__name__)

        return (single_qubit_coeffs, pair_qubit_coeffs, qubit_singles, qubit_pairs)

    @staticmethod
    def block_setter(
        input_object: Union[List[RotationGateMap], Hamiltonian], block_type: Enum
    ) -> List[RotationGateMap]:
        """
        Converts a Hamiltonian Object into a List of RotationGateMap Objects with
        the appropriate block_type and sequence assigned to the GateLabel

        OR

        Remaps a list of RotationGateMap Objects with a block_type and sequence
        implied from its position in the list.

        Parameters
        ----------
        input_object: `Union[List[RotationGateMap], Hamiltonian]`
            A Hamiltonian Object or a list of RotationGateMap Objects (Ordered
            according to their application order in the final circuit)
        block_type: Enum
            The type to be assigned to all the RotationGateMap Objects generated
            from input_object

        Returns
        -------
        `List[RotationGateMap]`
        """

        if isinstance(input_object, Hamiltonian):
            block = HamiltonianMapper.generate_gate_maps(input_object, block_type)
        elif isinstance(input_object, list):
            input_object = QAOADescriptor.set_block_sequence(input_object)
            for each_gate in input_object:
                if isinstance(each_gate, RotationGateMap):
                    each_gate.gate_label.update_gatelabel(new_gatemap_type=block_type)
                else:
                    raise TypeError(
                        f"Input gate is of unsupported type {type(each_gate)}."
                        "Only RotationGateMaps are supported"
                    )
            block = input_object
        else:
            raise ValueError(
                "The input object defining mixer should be a List of RotationGateMaps or type Hamiltonian"
            )
        return block

    @staticmethod
    def set_block_sequence(
        input_gatemap_list: List[RotationGateMap],
    ) -> List[RotationGateMap]:
        """
        This method assigns the sequence attribute to all RotationGateMap objects in the list.
        The sequence of the GateMaps are implied based on their positions in the list.

        Parameters
        ----------
        input_gatemap_list: `List[RotationGateMap]`
            A list of RotationGateMap Objects

        Returns
        -------
        `List[RotationGateMap]`
        """

        one_qubit_count = 0
        two_qubit_count = 0

        for each_gate in input_gatemap_list:
            if isinstance(each_gate, RotationGateMap):
                if each_gate.gate_label.n_qubits == 1:
                    each_gate.gate_label.update_gatelabel(
                        new_application_sequence=one_qubit_count,
                    )
                    one_qubit_count += 1
                elif each_gate.gate_label.n_qubits == 2:
                    each_gate.gate_label.update_gatelabel(
                        new_application_sequence=two_qubit_count,
                    )
                    two_qubit_count += 1
            else:
                raise TypeError(
                    f"Input gate is of unsupported type {type(each_gate)}."
                    "Only RotationGateMaps are supported"
                )
        return input_gatemap_list

    def reorder_gates_block(self, gates_block, layer_number):
        """Update the qubits that the gates are acting on after application
        of SWAPs in the cost layer
        """
        for gate in gates_block:
            if layer_number % 2 == 0:
                mapping = self.final_mapping
                gate.qubit_1 = mapping[gate.qubit_1]
                if gate.gate_label.n_qubits == 2:
                    gate.qubit_2 = mapping[gate.qubit_2]
            else:
                pass

        return gates_block

    @staticmethod
    def route_gates_list(
        gates_to_route: List[GateMap],
        device: "DeviceBase",
        routing_function: Callable,
    ) -> List[GateMap]:
        """
        Apply qubit routing to the abstract circuit gate list
        based on device information

        Parameters
        ----------
        gates_to_route: `List[GateMap]`
            The gates to route
        device: `DeviceBase`
            The device on which to run the circuit
        routing_function: `Callable`
            The function that accepts as input the device, problem, initial_mapping and
            outputs the list of gates with swaps
        """
        original_qubits_to_gate_mapping = {
            (gate.qubit_1, gate.qubit_2): gate
            for gate in gates_to_route
            if gate.gate_label.n_qubits == 2
        }
        problem_to_solve = list(original_qubits_to_gate_mapping.keys())
        (
            gate_list_indices,
            swap_mask,
            initial_physical_to_logical_mapping,
            final_mapping,
        ) = routing_function(device, problem_to_solve)

        gates_list = [gate for gate in gates_to_route if gate.gate_label.n_qubits == 1]
        swapped_history = []
        for idx, pair_ij in enumerate(gate_list_indices):
            mask = swap_mask[idx]
            qi, qj = pair_ij
            if mask == True:
                swapped_history.append(pair_ij)
                gates_list.append(SWAPGateMap(qi, qj))
            elif mask == False:
                old_qi, old_qj = qi, qj
                # traverse each SWAP application in reverse order to obtain
                # the original location of the current qubit
                for swap_pair in swapped_history[::-1]:
                    if old_qi in swap_pair:
                        old_qi = (
                            swap_pair[0] if swap_pair[1] == old_qi else swap_pair[1]
                        )
                    if old_qj in swap_pair:
                        old_qj = (
                            swap_pair[0] if swap_pair[1] == old_qj else swap_pair[1]
                        )
                try:
                    ising_gate = original_qubits_to_gate_mapping[
                        tuple([old_qi, old_qj])
                    ]
                except KeyError:
                    ising_gate = original_qubits_to_gate_mapping[
                        tuple([old_qj, old_qi])
                    ]
                except Exception as e:
                    raise e
                ising_gate.qubit_1, ising_gate.qubit_2 = qi, qj
                gates_list.append(ising_gate)

        return (
            gates_list,
            list(initial_physical_to_logical_mapping.keys()),
            final_mapping,
        )

    @property
    def abstract_circuit(self):
        # even layer inversion if the circuit contains SWAP gates
        even_layer_inversion = -1 if self.routed == True else 1
        _abstract_circuit = []
        for each_p in range(self.p):
            # apply each cost_block with reversed order to maintain the SWAP sequence
            _abstract_circuit.extend(
                self.cost_blocks[each_p][:: (even_layer_inversion) ** each_p]
            )
            # apply the mixer block
            if self.routed == True:
                mixer_block = self.reorder_gates_block(
                    self.mixer_blocks[each_p], each_p
                )
            else:
                mixer_block = self.mixer_blocks[each_p]
            _abstract_circuit.extend(mixer_block)

        return _abstract_circuit



import unittest
import networkx as nx
import numpy as np
from openqaoa.problems import MinimumVertexCover


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool

class TestMinimumVertexCover(unittest.TestCase):
    """Tests for MinimumVertexCover class"""
    def test_mvc_terms_weights_constant(self):
        """Test terms,weights,constant of QUBO generated by MVC class"""

        mvc_terms = [
            [0, 3],
            [0, 4],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
            [0],
            [1],
            [2],
            [3],
            [4],
        ]
        mvc_weights = [1.25, 1.25, 1.25, 1.25, 1.25, 1.25, 2.0, 2.0, 2.0, 3.25, 3.25]
        mvc_constant = 10.0

        gr = nx.generators.fast_gnp_random_graph(5, 0.8, seed=1234)
        mvc_prob = MinimumVertexCover(gr, field=1.0, penalty=5).qubo

        self.assertTrue(terms_list_equality(mvc_terms, mvc_prob.terms))
        self.assertEqual(mvc_weights, mvc_prob.weights)
        self.assertEqual(mvc_constant, mvc_prob.constant)

    def test_mvc_random_problem(self):
        """Test the random_instance method of MVC class"""
        mvc_terms = [
            [0, 3],
            [0, 4],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 4],
            [0],
            [1],
            [2],
            [3],
            [4],
        ]
        mvc_weights = [2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 4.5, 4.5, 4.5, 7.0, 7.0]
        mvc_constant = 17.5

        mvc_prob_random = MinimumVertexCover.random_instance(
            n_nodes=5, edge_probability=0.8, seed=1234
        ).qubo

        self.assertTrue(terms_list_equality(mvc_terms, mvc_prob_random.terms))
        self.assertEqual(mvc_weights, mvc_prob_random.weights)
        self.assertEqual(mvc_constant, mvc_prob_random.constant)

    def test_mvc_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {"node1": 1, "node2": 2}, np.array([1, 2])]
        field = 0.1
        penalty = 0.1

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(each_graph, field, penalty)
            self.assertEqual(
                "Input problem graph must be a networkx Graph.", str(e.exception)
            )

        # field capacity type-check
        graph = nx.circulant_graph(6, [1])
        field_list = [np.array(1), np.array(0.5), "oh"]
        penalty = 0.1

        for each_field in field_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(graph, each_field, penalty)
            self.assertEqual(
                "The input parameter, field, has to be of type float or int",
                str(e.exception),
            )

        # penalty capacity type-check
        graph = nx.circulant_graph(6, [1])
        field = 0.1
        penalty_list = [np.array(1), np.array(0.5), "oh"]

        for each_penalty in penalty_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(graph, field, each_penalty)
            self.assertEqual(
                "The input parameter, penalty, has to be of type float or int",
                str(e.exception),
            )


if __name__ == "__main__":
    unittest.main()



from braket.jobs import save_job_result

import openqaoa
from openqaoa_braket.algorithms import AWSJobs


def main():
    """
    The entry point is kept clean and simple and all the load statements are hidden in the `aws_jobs_load` function (which will become part of the OpenQAOA library)
    """

    job = AWSJobs(algorithm="QAOA")
    job.load_compile_data()
    job.run_workflow()

    save_job_result({"result": job.workflow.asdict()})


if __name__ == "__main__":
    main()



import numpy as np

from ...qaoa_components import Hamiltonian
from ...utilities import bitstring_energy
from ...problems import QUBO


def max_terms(exp_vals_z: np.ndarray, corr_matrix: np.ndarray, n_elim: int):
    """
    Extracts the n_elim expectation values (single spin and correlation) with
    highest magnitude, and uses them to impose the elimination constraint on
    the spins.

    Parameters
    ----------
    exp_vals_z: `np.array`
        Single spin expectation values.
    corr_matrix: `np.array`
        Correlation matrix.
    n_elim: `int`
        Number of spins to eliminate.

    Returns
    -------
    max_terms_and_stats: `dict`
        Dictionary containing terms to be eliminated and their expectation values.
    """
    # Copy list of single spin expectation values
    Z = exp_vals_z.copy()

    # Copy correlation matrix
    M = corr_matrix.copy()

    # Initialize dict of terms and respective costs
    max_terms_and_stats = {}

    # Find the n_max+1 largest correlation pairs
    for _ in range(n_elim):
        # Obtain max_spin, its Z expectation value, and store it - round to
        # suppress errors from numbers approximating 0
        max_spin = np.unravel_index(abs(Z).argmax(), Z.shape)
        max_val = np.round(Z[max_spin], 10)

        # Obtain max_term, its correlation, and store it
        max_term = np.unravel_index(abs(M).argmax(), M.shape)
        max_corr = np.round(M[max_term], 10)

        # Eliminate the expectation value with highest absolute value - round to
        # suppress errors from numbers approximating 0
        if np.abs(max_corr) > np.abs(max_val):
            max_terms_and_stats.update({max_term: max_corr})

            # Set the term to 0 to check for the next one
            M[max_term] = 0

        else:
            # If maximum value is 0, all expectation values are vanishing
            if max_val == 0:
                break
            else:
                max_terms_and_stats.update({max_spin: max_val})

            # Set the spin value to 0 to check for the next one
            Z[max_spin] = 0

    # Flag if we have have not been able to extract any relation for the terms
    if max_terms_and_stats == {}:
        print("All expectation values are 0: Breaking degeneracy by fixing a qubit\n")
        max_terms_and_stats = {(0,): -1.0}

    return max_terms_and_stats


def ada_max_terms(exp_vals_z: np.ndarray, corr_matrix: np.ndarray, n_max: int):
    """
    Extracts the n_max+1 expectation values (single spin and correlation) with
    highest magnitude, computes the average among them and selects the ones
    above average for elimination. The maximum number of potential candidates is n_max.

    Parameters
    ----------
    exp_vals_z: `np.array`
        Single spin expectation values.
    corr_matrix: `np.array`
        Correlation matrix.
    n_max: `np.array`
        Maximum number of potential candidates for elimination.

    Returns
    -------
    max_terms_and_stats: `dict`
        Dictionary containing terms to be eliminated and their expectation values.
    """

    # Copy list of single spin expectation values
    Z = exp_vals_z.copy()

    # Copy correlation matrix
    M = corr_matrix.copy()

    # Initialize dict of terms and respective costs
    max_terms_and_stats = {}

    # Find the n_max+1 largest correlation pairs
    for _ in range(n_max + 1):
        # Obtain max_spin, its Z expectation value, and store it - round to
        # suppress errors from numbers approximating 0
        max_spin = np.unravel_index(abs(Z).argmax(), Z.shape)
        max_val = np.round(Z[max_spin], 10)

        # Obtain max_term, its correlation, and store it
        max_term = np.unravel_index(abs(M).argmax(), M.shape)
        max_corr = np.round(M[max_term], 10)

        # Eliminate the expectation value with highest absolute value - round to
        # suppress errors from numbers approximating 0
        if np.abs(max_corr) > np.abs(max_val):
            max_terms_and_stats.update({max_term: max_corr})

            # Set the term to 0 to check for the next one
            M[max_term] = 0

        else:
            # If maximum value is 0, all expectation values are vanishing
            if max_val == 0:
                break
            else:
                max_terms_and_stats.update({max_spin: max_val})

            # Set the spin value to 0 to check for the next one
            Z[max_spin] = 0

    # Flag if we have have not been able to extract any relation for the terms
    if max_terms_and_stats == {}:
        print("All expectation values are 0: Breaking degeneracy by fixing a qubit\n")
        max_terms_and_stats = {(0,): -1.0}

    # Correlation average magnitude
    avg_mag_stats = np.round(np.mean(np.abs(list(max_terms_and_stats.values()))), 10)

    # Select only the ones above average
    max_terms_and_stats = {
        key: value
        for key, value in max_terms_and_stats.items()
        if np.abs(value) >= avg_mag_stats
    }

    # Cut down the number of eliminations if, due to symmetry, they exceed
    # the number allowed - relevant for unweighted graphs
    if len(max_terms_and_stats) > n_max:
        max_keys = list(max_terms_and_stats.keys())[0:n_max]
        max_terms_and_stats = {key: max_terms_and_stats[key] for key in max_keys}

    return max_terms_and_stats


def find_parent(spin_map: dict, spin: int, factor: int = 1):
    """
    Finds parent spin recursively following the chain of dependencies in the spin map.

    Parameters
    ----------
    spin_map: `dict`
        Mapping containing all dependencies for eliminated spins.
    spin: `int`
        Spin whose parent we want to find.
    factor: `int`, optional
        Defaults to 1 to initialize the multiplicative factor.

    Returns
    -------
    parent_spin: `int`
        Parent spin.
    factor: `int`
        Cumulative factor connecting the input spin with its parent spin.
    """

    # Extract parent spin
    parent_spin = spin_map[spin][1]
    factor *= spin_map[spin][0]

    # If spin maps to itself or it is None, we have found parent spin
    if parent_spin == spin or parent_spin is None:
        return parent_spin, factor

    # Recursively follow dependencies
    else:
        spin = parent_spin

    return find_parent(spin_map, spin, factor)


def spin_mapping(problem: QUBO, max_terms_and_stats: dict):
    """
    Generates a map between spins in the original problem graph and in the reduced graph.
    Elimination constraints from correlations define a constrained spin, to be removed, and a
    parent spin, to be kept. Parent spins determine the state of multiple spins by a
    chain of dependencies between spins due to the different constraints. Note that there
    is always a parent spin and only one. If cycles are present, less edges will
    be eliminated to satisfy this requirement. Constraints following from biases result in
    fixing spins to a specific value. In this case, the parent spin is set
    to None. Spins in the map that are not eliminated are mapped to themselves.

    Parameters
    ----------
    problem: `QUBO`
        QUBO problem object containing the problem statement.
    max_terms_and_stats: `dict`
        Dictionary containing the terms to be eliminated and their expectation values.

    Returns
    -------
    spin_map: `dict`
        Dictionary containing all the mapping dependencies. The keys
        correspond to the original spins. The values are tuple pairs,
        containing the parent spin and the factor relating both.
        The structure is: { old_spin : (factor, new_spin) }.
    """

    register = list(range(problem.n))
    max_terms = list(max_terms_and_stats.keys())

    # Initialize with every spin mapping to itself
    spin_map = dict((spin, (1, spin)) for spin in register)

    # Retrieve all spins involved in terms with highest correlation
    spin_candidates = set([spin for term in max_terms for spin in term])

    # Order term entries in descending magnitude order for correct insertion in solution
    sorted_max_ts = sorted(
        max_terms_and_stats.items(), key=lambda x: np.abs(x[1]), reverse=True
    )

    # Build spin map from all expectation values
    for term, stat in sorted_max_ts:
        # Single spin expectation value
        if len(term) == 1:
            # Select spin
            spin = term[0]

            # Check if it has been fixed already
            parent, _ = find_parent(spin_map, spin)

            # If not, fix it
            if parent is not None:
                spin_map.update({spin: (np.sign(stat), None)})

        # Correlation terms
        else:
            # Select spins
            spin_keep, spin_remove = term

            # Retrieve parents and fact
            parent_keep, factor_keep = find_parent(spin_map, spin_keep)
            parent_remove, factor_remove = find_parent(spin_map, spin_remove)

            # If neither spin has been fixed, store relation
            if parent_keep is not None and parent_remove is not None:
                # Ensure we keep the spin with lowest label for future state reconstruction
                if parent_remove < parent_keep:
                    parent_keep, parent_remove = parent_remove, parent_keep
                    factor_keep, factor_remove = factor_remove, factor_keep

                # If a cycle is encountered one edge is discarded
                if parent_remove == parent_keep:
                    continue

                # Update the spin map
                else:
                    spin_map.update(
                        {
                            parent_remove: (
                                factor_remove ** (-1) * factor_keep * np.sign(stat),
                                parent_keep,
                            )
                        }
                    )

            # If both spins have been fixed, ignore correlation
            elif parent_keep is None and parent_remove is None:
                continue

            # If one spin has been fixed, fix the second one according to correlation value
            else:
                # Extract fixed and unfixed spins
                spin_fixed, factor_fixed = (
                    (parent_keep, factor_keep)
                    if parent_keep is None
                    else (parent_remove, factor_remove)
                )
                spin_unfixed, factor_unfixed = (
                    (parent_remove, factor_remove)
                    if spin_fixed == parent_keep
                    else (parent_keep, factor_fixed)
                )

                # Fix spin
                spin_map.update(
                    {
                        spin_unfixed: (
                            factor_unfixed ** (-1) * factor_fixed * np.sign(stat),
                            spin_fixed,
                        )
                    }
                )

    # Correct all dependencies
    for spin in spin_candidates:
        parent_spin, cumulative_factor = find_parent(spin_map, spin)
        spin_map.update({spin: (cumulative_factor, parent_spin)})

    return spin_map


def problem_from_dict(problem_dict: dict):
    """
    Transforms a QUBO problem, input as a dictionary, into a QUBO problem, output as
    a QUBO object, ensuring proper labelling of the nodes. For example, for a set
    of nodes [0,1,4,6] with edges [(0,1),(1,4),(4,6),(0,6)], after the relabelling, the
    Hamiltonian object will be constructed with node labels [0,1,2,3] and edges
    [(0,1),(1,2),(2,3),(1,3)].

    Parameters
    ----------
    problem_dict: `dict`
        QUBO problem as a dictionary containing the edges as keys and weights as values.

    Returns
    -------
    problem: `QUBO`
        A QUBO problem object constructed using the classical_hamiltonian() method.
    """

    edges = list(problem_dict.keys())
    weights = list(problem_dict.values())

    # Current indices after elimination
    register = list(set([i for edge in edges for i in edge]))

    # Index mapping
    label_mapping = {unique: i for i, unique in enumerate(register)}

    # Initialize qubit terms mapper
    label_edges_mapping = {}

    # Edges mapping
    for edge in edges:
        # Map linear term to linear term
        if len(edge) == 1:
            label_edges_mapping.update({edge: (label_mapping.get(edge[0]),)})

        # Map quadratic term to quadratic term
        elif len(edge) == 2:
            label_edges_mapping.update(
                {edge: (label_mapping.get(edge[0]), label_mapping.get(edge[1]))}
            )

        # If constant term, just map to itself
        else:
            label_edges_mapping.update({edge: edge})

    # New edges
    new_edges = list(label_edges_mapping.values())

    # New hamiltonian
    problem = QUBO(n=len(register), terms=new_edges, weights=weights)

    return problem


def redefine_problem(problem: QUBO, spin_map: dict):
    """
    Returns the QUBO object of the reduced problem. Using the spin map, we construct a  dictionary
    containing the new set of edges and weights defining the new problem.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.
    spin_map: `dict`
        Dictionary containing the spin dependencies.

    Returns
    -------
    new_problem: `QUBO`
        QUBO object containing the reduced problem statement after spin eliminations.
    spin_map: `dict`
        Updated spin_map with sponatenous eliminations from cancellations during spin removal process.
    """

    # Define new QUBO problem as a dictionary
    old_register = set([spin for term in problem.terms for spin in term])
    new_problem_dict = {}

    # get a set of all the spins to be eliminated
    eliminated_spins = set()
    for spin in spin_map.keys():
        if spin != spin_map[spin][1] or spin_map[spin][1] is None:
            eliminated_spins.add(spin)

    # Scan all terms and weights
    for term, weight in zip(problem.terms, problem.weights):
        # Bias terms
        if len(term) == 1:
            # Extract spin from the map
            spin = term[0]

            # Extract parent spin and associated factor
            parent_spin, factor_spin = spin_map[spin][1], spin_map[spin][0]

            # If spin is fixed it only contributes with a constant term
            if parent_spin is None:
                continue

            # If unfixed, define new edge and weight
            else:
                new_edge = (parent_spin,)
                new_weight = factor_spin * weight

                # Add new edge if not already present in the dictionary
                if new_problem_dict.get(new_edge) is None:
                    new_problem_dict.update({new_edge: new_weight})

                # If new edge already present, add weight
                else:
                    new_problem_dict[new_edge] += new_weight

        # Quadratic terms
        else:
            # Extract spins from term
            spin1, spin2 = term

            # Extract parent spins and associated factors
            factor_spin1, factor_spin2 = spin_map[spin1][0], spin_map[spin2][0]
            parent_spin1, parent_spin2 = spin_map[spin1][1], spin_map[spin2][1]

            # If both spins are fixed or have same parent, they only contribute with a constant term
            if parent_spin1 == parent_spin2:
                continue

            # If neither parent spin is fixed and they are different, define new quadratic term
            elif parent_spin1 is not None and parent_spin2 is not None:
                # Define new edge through the spin map and order them
                new_edge = [parent_spin1, parent_spin2]
                new_edge = tuple([min(new_edge), max(new_edge)])

                # Define new weight from factors in the spin map
                new_weight = factor_spin1 * factor_spin2 * weight

                # Add new edge if not already present in the dictionary
                if new_problem_dict.get(new_edge) is None:
                    new_problem_dict.update({new_edge: new_weight})

                # If new edge already present, add weight
                else:
                    new_problem_dict[new_edge] += new_weight

            # If only one spin is fixed, quadratic term becomes a bias term
            else:
                # Define new bias term keeping the unfixed spin
                new_edge = (parent_spin1,) if parent_spin2 is None else (parent_spin2,)
                new_weight = factor_spin1 * factor_spin2 * weight

                # Add new edge if not already present in the dictionary
                if new_problem_dict.get(new_edge) is None:
                    new_problem_dict.update({new_edge: new_weight})

                # If new term already present, add weight
                else:
                    new_problem_dict[new_edge] += new_weight

    # New qubit register
    new_register = set([spin for term in new_problem_dict.keys() for spin in term])

    # Remove vanishing edges
    new_problem_dict = {
        edge: weight
        for edge, weight in new_problem_dict.items()
        if round(weight, 10) != 0
    }

    # Define quadratic register after removing vanishing terms
    new_quadratic_register = set(
        [spin for edge in new_problem_dict.keys() if len(edge) == 2 for spin in edge]
    )

    # Check for isolated nodes after constructing the new problem
    there_is_isolated_nodes = False

    # If lengths do not match, there are isolated nodes
    if len(new_register) != len(new_quadratic_register):
        isolated_nodes = new_register.difference(new_quadratic_register)
        there_is_isolated_nodes = True
    # If too few eliminations, there are isolated nodes; only important for bias-free problems.
    elif old_register - new_register != eliminated_spins:
        isolated_nodes = old_register.difference(new_register)
        isolated_nodes = isolated_nodes.difference(eliminated_spins)
        there_is_isolated_nodes = True

    if there_is_isolated_nodes:
        # Fix isolated nodes
        for node in isolated_nodes:
            singlet = (node,)

            # If no linear term acting on the node, fix arbitrarily
            if new_problem_dict.get(singlet) is None:
                spin_map.update({node: (1, None)})

            # If linear term present, fix accordingly by anti-aligning
            else:
                factor = -np.sign(new_problem_dict.get(singlet))
                spin_map.update({node: (factor, None)})

                # Delete isolated node from new problem
                new_problem_dict.pop((node,))

    # For some unweighted graphs specific eliminations can lead to eliminating
    # the whole instance before reaching cutoff.
    if new_problem_dict == {}:
        # set the problem to the old problem and solveclassically for the smallest non-vanishing instance.
        new_problem = problem

    else:
        # Redefine new QUBO problem from the dictionary
        new_problem = problem_from_dict(new_problem_dict)

    return new_problem, spin_map


def final_solution(
    elimination_tracker: list, cl_states: list, hamiltonian: Hamiltonian
):
    """
    Constructs the final solution to the problem by obtaining the final states
    from adding the removed spins into the classical results and
    computing the corresponding energy.

    Parameters
    ----------
    elimination_tracker: `list`
        List of dictionaries, where each dictionary contains the elimination rules
        applied at each step of the process. Dictionary keys correspond to spin
        pairs (i,j), always with i<j to ensure proper reconstruction of the state,
        where spin j was eliminated in favor of i. The values of each pair correspond
        to the correlation sign between spins i and j. For fixed spins, in the pair (i,j),
        i is None, and the dictionary value corresponds to the fixed state of j.
    cl_states: `list`
        Set of states as solutions in
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the original problem statement.

    Returns
    -------
    full_solution: `dict`
        Dictionary containing the solution states of the problem and their
        respective energies.
    """

    # Reverse max_cost_list, reverse max_pair_list
    elimination_tracker = elimination_tracker[::-1]

    # Initialize list containg full solutions
    full_solution = {}

    # Re-construct each solution
    for cl_state in cl_states:
        # Transform every string solution into binary integer
        state = [int(bit) for bit in cl_state]

        # Extract terms and costs from each elimination step
        for terms_and_stats in elimination_tracker:
            # Back track elimination from the specific step
            for term_stat in terms_and_stats:
                if "singlet" in term_stat:
                    (i,) = term_stat["singlet"]
                    val = term_stat["bias"]
                    val = term_stat["bias"]

                    # Basis change
                    binary_val = int((1 - val) / 2)

                    # Insert fixed value of i
                    state.insert(i, binary_val)

                # If we have a pair
                else:
                    i, j = term_stat["pair"]
                    val = term_stat["correlation"]
                    val = term_stat["correlation"]

                    # Insert new value in position j according to correlation with i
                    prev_corr = state[i]

                    state.insert(j, prev_corr) if val > 0 else state.insert(
                        j, prev_corr ^ 1
                    )

        # Store solution states and their energy
        full_solution.update(
            {"".join(str(i) for i in state): bitstring_energy(hamiltonian, state)}
        )

    return full_solution


def solution_for_vanishing_instances(hamiltonian: Hamiltonian, spin_map: dict):
    """
    Constructs the final solution of the smallest non vanishing problem by fixing the vanishing spins arbitrarily to 1
    while obeying the correlations identified by the last run of QAOA before the problem vanished.
    Computing the classical energy of the generated string.

    Parameters
    ----------
    spin_map: `dict`
        Spin map containing the correlations and eliminations of the smallest non-vanishing
        problem statement.
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the smallest non vanishing problem statement.

    Returns
    -------
    cl_energy: `float`
        The energy of the solution wrt the cost Hamiltonian.
    cl_ground_state: `list`
        The (single) classical solution to the problem reconstructed accordingly to the last spin map.
        Represented as a binary string and then cast to a list to match the output of the
        `ground_state_hamiltonian` function which is used usually.
    """
    cl_ground_state = ""

    for spin in spin_map.keys():
        if spin_map[spin][1] is None:
            cl_ground_state += "1"
        else:
            # fix according to correlation factor
            factor = spin_map[spin][0]
            parent = spin_map[spin][1]

            if factor == 1.0:
                # correlated
                cl_ground_state += cl_ground_state[parent]
            else:
                # anticorrelated
                cl_ground_state += str(int(not bool(int(cl_ground_state[parent]))))

    # computing the energy of the first one only, assuming degeneracy
    cl_energy = bitstring_energy(hamiltonian, cl_ground_state)

    # convert to list to match the expected format
    cl_ground_state = [cl_ground_state]

    return cl_energy, cl_ground_state
